VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "aeSocket"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = True
Attribute VB_Ext_KEY = "SavedWithClassBuilder6" ,"Yes"
Attribute VB_Ext_KEY = "Top_Level" ,"Yes"
Option Explicit

'AzUrE
'Copyright (C) 2000 Sam Smith All Rights Reserved
'
'Contact: <samuel.smith@nme.com>
'Version: 1.0.0.0
'
'This software is distributed in the hope that it will be useful,
'but WITHOUT ANY WARRANTY; without even the implied warranty of
'MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

Public Enum aeSocketProtocol
  'Use TCP stream
  sck_TCP = 1
  'Use UDP datagrams
  sck_UDP = 2
End Enum

Public Enum aeSocketSendOptions
  'Send normal
  SendNormal = 0
  'Send out of band
  SendOOB = 1
  'send without using routing tables
  SendNoRoute = 2
End Enum

Public Enum aeSOCKSProtocol
  'Use SOCKS4 protocol:
  SOCKS_V4 = 1
  'Use SOCKS5 protocol
  SOCKS_V5 = 2
End Enum

Public Enum aeSOCKSMethod
  'Connect through
  SOCKSConnect = 1
  'Bind on
  SOCKSBind = 2
End Enum

Private Type SOCKSData
  'Use SOCKS or not
  bUseSOCKS As Boolean
  'Current stage
  Level As Long
  'Which protocol to use
  TOC As aeSOCKSProtocol
  'Connect or bind
  Meth As aeSOCKSMethod
  'Proxy hostname or IP
  ProxyName As String
  'IP address of 'ProxyName' maybe the same as ProxyName
  ProxyIP As String
  'Proxy port, usually 1080
  ProxyPort As Long
  'IP address of host to connect to (must be IP)
  ThatIP As String
  'Port at IP address to connect on
  ThatPort As Long
  'Amount in seconds to allow point of connection to proxy
  'to 2nd connection or bind event
  NegTimeOut As Long
  'Username used for authentication
  UserName As String
  'Password used for authentication
  PassWord As String
  'Are both Username and Password present?
  bUP As Boolean
  'IP address that has the binded port
  BindHost As String
  'Port SOCKS has binded for us
  lBindPort As Long
  'IP address the 3rd machine see's us as
  SeenIP As String
  'The local port on the SOCKS server that the 3rd host see's as it's remote port
  SeenPort As Long
  'Host/IP of the computer we are connected to from a BINDed socket on SOCKS
  IndirectHost As String
  'Port
  IndirectPort As Long
End Type

'local copy of SOCKSdata
Private hS As SOCKSData

Private Type sckOpt
  'Number of connections to backlog
  lAcceptQ As Long
  'Are keep alive packets being sent?
  bKeepAlive As Boolean
  'Linger on close
  bLinger As Boolean
  'Is OOB data sent inline with the normal stream
  bBandInLine As Boolean
  'Can addr be bound by others?
  bReUseAddr As Boolean
  'Dont use routing tables
  bDontRoute As Boolean
  'Size of get buffer
  lGetBufLen As Long
  'Size of send buffer
  lSendBufLen As Long
  'Timeout for a recv()
  lGetTimeOut As Long
  'Timeout for send()
  lSendTimeOut As Long
  'Dont use the Nagle algorithm?
  bTCPNoDelay As Boolean
  'Allow the sending of broadcast messages (UDP only)
  bAllowBroadcast As Boolean
End Type

Private hSck As sckOpt

Private Type wskProp
  'Pointer to this class
  hClassPtr As Long
  'Current pointer to event interface
  'can be internal or external
  hEventPtr As Long
  'Current socket handle
  hSocket As Long
  'Current state of the socket,
  'we must keep it "up to date"
  hState As aeSocketState
  'Current protocol of our socket, TCP and UDP are supported
  'can be "switched" at any time, though access to this
  'property is read only, should only be changed at ONE access point
  hProtocol As aeSocketProtocol
  'String containing remote IP if any, all checking of this string
  'should be done by the owning property
  RemotePeer As String
  'String containing IP address on iface used
  'to make the connection
  LocalPeer As String
  'Local port
  LocalPort As Long
  'Remote port on remote peer
  RemotePort As Long
  'struct containing, long Ip, port, address family
  sck_in As sockaddr_in
  'should be read only to the client
  RemoteIP As String
  'socket can be written to
  sckWrite As Boolean
  'socket can be read from
  sckRead As Boolean
  'how much the socket has sent in bytes
  sntB As Long
  'how much read from socket?
  rcvB As Long
  'Reuse socket options
  bSocketReuse As Boolean
End Type

'local copy of our udt
Private hData As wskProp

Private Type MemoryBuffer
  'Handle to current memory object
  hMem As Long
  'Size in bytes of used memory
  Size As Long
  'Size in bytes of memory object (what GlobalSize would return)
  ActualTotalSize As Long
  'Point to valid write to maybe -1
  WritePoint As Long
  'If True don't call Lock/Unlock
  DontWrite As Boolean
  '
  IsWriting As Boolean
End Type

'local copy of udt
Private hBuf As MemoryBuffer

'Number of seconds to wait, before stopping
Private mvarTimeout As Long

'timer attached flag
Private mvarTimerRunning As Boolean

'Use a buffer or disgard unread data
Private mvarUseBuffer As Boolean

'Fire events in implemented interface(if the internal one is used)
Private mvarUseEvents As Boolean

'Count bytes for send/recv calls
Private mvarUseByteCount As Boolean

'DNS resolves needed for name look up
Private cDNS As aeDNS

'Interface needed to get results from DNS class
Implements aeDNSNotify

'Need subclass for winsock messages and timer
Implements aeSubclass

'Event interface
Implements aeSocketEvent

'Local copy needed to be kept in private global memory
Private hPack As HostPortData

'Events, these are only raised by our internal
'implemented interface and can not be raised externally
'Any code that uses the events should be careful
'not to use 'UseEvents = False' or to set a new interface

Public Event Opened(RemotePeer As String, RemotePort As Long, _
  LocalPeer As String, LocalPort As Long, _
  bFromListen As Boolean)
Public Event Closed( _
  RemotePeer As String, RemotePort As Long, bRemoteClosure As Boolean, bGrace As Boolean, bKeptData As Boolean)
Public Event Error( _
  ErrorCode As Long, ErrorString As String)
Public Event AcceptConnection(RemotePeer As String, _
  RemotePort As Long, SocketID As Long)
Public Event DNSLookUpDone(Host As String, bIsHost As Boolean, _
  PassedHost As String, bError As Boolean, _
  ErrorCode As Long, ErrorCodeMessage As String)
Public Event DNSLookUpStarted( _
  Host As String, bIsHost As Boolean)
Public Event StateChanged( _
  NewState As aeSocketState)
Public Event SocketChanged( _
  NewHandle As Long, SelfPtr As Long)
Public Event DataRead( _
  ByVal BufferLen As Long, _
  ByVal Buffered As aeSocketReadDataEnum, _
  ByVal DataPtr As Long, _
  Data() As Byte, _
  ByVal IsOOB As Boolean)
Public Event Binded( _
  ByVal Port As Long, _
  ByVal bRemote As Boolean)
  
'Errors defined, no "hard" error are raised
'The error base can be changed, errors refer to the consts
'and not the actual number
Private Const aeWinsockError_Base = 10350
Private Const aeWinsockError_BadHost = (aeWinsockError_Base + 1)
Private Const aeWinsockError_BadPort = (aeWinsockError_Base + 2)
Private Const aeWinsockError_BadState2Send = (aeWinsockError_Base + 3)
Private Const aeWinsockError_CantUseDataType = (aeWinsockError_Base + 4)
Private Const aeWinsockError_CantSendEmptyString = (aeWinsockError_Base + 5)
Private Const aeWinsockError_FailSendInternal = (aeWinsockError_Base + 6)
Private Const aeWinsockError_BadArray = (aeWinsockError_Base + 7)
Private Const aeWinsockError_BadMemory = (aeWinsockError_Base + 8)
Private Const aeWinsockError_NoMemory = (aeWinsockError_Base + 9)
Private Const aeWinsockError_NoPtr = (aeWinsockError_Base + 10)
Private Const aeWinsockError_BadPortArgumentMinus = (aeWinsockError_Base + 11)
Private Const aeWinsockError_AcceptSocketInvalid = (aeWinsockError_Base + 12)
Private Const aeWinsockError_NoProxyResolve = (aeWinsockError_Base + 14)
Private Const aeWinsockError_NoConnectProxy = (aeWinsockError_Base + 15)
Private Const aeWinsockError_ProxyConnTimeOut = (aeWinsockError_Base + 16)
Private Const aeWinsockError_NegTimeOut = (aeWinsockError_Base + 17)
Private Const aeWinsockError_CantSendPacketV4 = (aeWinsockError_Base + 18)
Private Const aeWinsockError_NotSOCKS4 = (aeWinsockError_Base + 19)
Private Const aeWinsockError_CantReadPacketV4 = (aeWinsockError_Base + 20)
Private Const aeWinsockError_NoErrorReplyV4 = (aeWinsockError_Base + 21)
Private Const aeWinsockError_NotSOCKS4Reply = (aeWinsockError_Base + 22)
Private Const aeWinsockError_ErrorUnknownV4 = (aeWinsockError_Base + 23)
Private Const aeWinsockError_RejectedOrFailV4 = (aeWinsockError_Base + 24)
Private Const aeWinsockError_CantConnectToIdentd = (aeWinsockError_Base + 25)
Private Const aeWinsockError_UserNameIdentNonMatch = (aeWinsockError_Base + 26)
Private Const aeWinsockError_InvalidPortV4 = (aeWinsockError_Base + 27)
Private Const aeWinsockError_CloseAfterBindV4 = (aeWinsockError_Base + 28)
Private Const aeWinsockError_CantSendMethodReqV5 = (aeWinsockError_Base + 29)
Private Const aeWinsockError_BadReplyForMethodV5 = (aeWinsockError_Base + 30)
Private Const aeWinsockError_BadReadV5 = (aeWinsockError_Base + 31)
Private Const aeWinsockError_BadMethodReplyV5 = (aeWinsockError_Base + 32)
Private Const aeWinsockError_NoMethodsV5 = (aeWinsockError_Base + 33)
Private Const aeWinsockError_UnknownMethodV5 = (aeWinsockError_Base + 34)
Private Const aeWinsockError_BadSendUPV5 = (aeWinsockError_Base + 35)
Private Const aeWinsockError_UPReplyTooBig = (aeWinsockError_Base + 36)
Private Const aeWinsockError_BadUPVersionAuth = (aeWinsockError_Base + 37)
Private Const aeWinsockError_BadPasswordV5 = (aeWinsockError_Base + 38)
Private Const aeWinsockError_BadSendOfHostInfo = (aeWinsockError_Base + 39)
Private Const aeWinsockError_SOCKS5ReplyToSmall = (aeWinsockError_Base + 40)
Private Const aeWinsockError_SOCKS5VersioninReplyInvalid = (aeWinsockError_Base + 41)
Private Const aeWinsockError_SOCKS5GeneralFail = (aeWinsockError_Base + 42)
Private Const aeWinsockError_SOCKS5Rulesetfail = (aeWinsockError_Base + 43)
Private Const aeWinsockError_SOCKS5NetUnreach = (aeWinsockError_Base + 44)
Private Const aeWinsockError_SOCKS5HostUnreach = (aeWinsockError_Base + 45)
Private Const aeWinsockError_SOCKS5ConnectionRefused = (aeWinsockError_Base + 46)
Private Const aeWinsockError_SOCKS5TTLExpired = (aeWinsockError_Base + 47)
Private Const aeWinsockError_SOCKS5ComNoSupport = (aeWinsockError_Base + 48)
Private Const aeWinsockError_SOCKS5AddrTypeNotSupported = (aeWinsockError_Base + 49)
Private Const aeWinsockError_SOCKS5UnAssignedError = (aeWinsockError_Base + 50)
Private Const aeWinsockError_SOCKS5BadReplyNoRes = (aeWinsockError_Base + 51)
Private Const aeWinsockError_SOCKSRepAddrTypeNoSupportByClient = (aeWinsockError_Base + 52)
Private Const aeWinsockError_SOCKSCloseLevel1 = (aeWinsockError_Base + 53)
Private Const aeWinsockError_SOCKSCloseLevel2 = (aeWinsockError_Base + 54)
Private Const aeWinsockError_SOCKSCloseLevel3 = (aeWinsockError_Base + 55)
Private Const aeWinsockError_SOCKSCloseLevel4 = (aeWinsockError_Base + 56)
Private Const aeWinsockError_SOCKSCloseLevel5 = (aeWinsockError_Base + 57)

Private Function gErrStr(ByVal ErrorCode As Long) As String
  Select Case ErrorCode
    Case 0
      gErrStr = "No Error"
    Case aeWinsockError_BadHost
      gErrStr = "No host was passed"
    Case aeWinsockError_BadPort
      gErrStr = "No port was passed"
    Case aeWinsockError_BadState2Send
      gErrStr = "Can not send data, socket was not open or connected"
    Case aeWinsockError_CantUseDataType
      gErrStr = "Can not use data type"
    Case aeWinsockError_CantSendEmptyString
      gErrStr = "Can not send an empty string"
    Case aeWinsockError_FailSendInternal
      gErrStr = "There was an internal error and data could not be sent"
    Case aeWinsockError_BadMemory
      gErrStr = "Memory block pointer passed was null or invalid"
    Case aeWinsockError_NoMemory
      gErrStr = "Memory block contained no data or was invalid"
    Case aeWinsockError_NoPtr
      gErrStr = "Could not obtain pointer from memory object (lock failed)"
    Case aeWinsockError_BadPortArgumentMinus
      gErrStr = "Can not use passed port number, less than zero"
    Case aeWinsockError_AcceptSocketInvalid
      gErrStr = "Socket ID passed to wsAccept was not valid or could not be used"
    Case aeWinsockError_NoProxyResolve
      gErrStr = "Unable to resolve proxy hostname"
    Case aeWinsockError_NoConnectProxy
      gErrStr = "Unable to connect to proxy host"
    Case aeWinsockError_ProxyConnTimeOut
      gErrStr = "Connection to proxy host timed out"
    Case aeWinsockError_NegTimeOut
      gErrStr = "Allowed time for sub-negotiation/connection has been reached"
    Case aeWinsockError_CantSendPacketV4
      gErrStr = "Could not send SOCKS4 connection/bind packet"
    Case aeWinsockError_NotSOCKS4
      gErrStr = "Remote server did not support SOCKS4 protocol"
    Case aeWinsockError_CantReadPacketV4
      gErrStr = "Could not read sent packet from SOCKS4 server"
    Case aeWinsockError_NoErrorReplyV4
      gErrStr = "SOCKS4 closed the connection without sending error information"
    Case aeWinsockError_NotSOCKS4Reply
      gErrStr = "Reply packet was invalid"
    Case aeWinsockError_ErrorUnknownV4
      gErrStr = "SOCKS4 replied with unknown failure"
    Case aeWinsockError_RejectedOrFailV4
      gErrStr = "SOCKS4 rejected request/reported failure"
    Case aeWinsockError_CantConnectToIdentd
      gErrStr = "SOCKS4 could not connect to IDENTD"
    Case aeWinsockError_UserNameIdentNonMatch
      gErrStr = "IdentD username and given username did not match"
    Case aeWinsockError_InvalidPortV4
      gErrStr = "SOCKS4 returned an invalid port"
    Case aeWinsockError_CloseAfterBindV4
      gErrStr = "SOCKS4 closed the binded port and connection"
    Case aeWinsockError_CantSendMethodReqV5
      gErrStr = "Could not send SOCKS5 method request"
    Case aeWinsockError_BadReplyForMethodV5
      gErrStr = "Remote server did not support SOCKS5 protocol"
    Case aeWinsockError_BadReadV5
      gErrStr = "Could not read SOCKS5 packet from server"
    Case aeWinsockError_BadMethodReplyV5
      gErrStr = "SOCKS5 method packet was invalid"
    Case aeWinsockError_NoMethodsV5
      gErrStr = "SOCKS5 did not support any requested methods"
    Case aeWinsockError_UnknownMethodV5
      gErrStr = "SOCKS replied with an unknown method"
    Case aeWinsockError_BadSendUPV5
      gErrStr = "Could not send username/password authentication packet"
    Case aeWinsockError_UPReplyTooBig
      gErrStr = "Username/password packet reply size was invalid"
    Case aeWinsockError_BadUPVersionAuth
      gErrStr = "Username/Password authentication reply packet version was invalid"
    Case aeWinsockError_BadPasswordV5
      gErrStr = "Authentication of username/password failed"
    Case aeWinsockError_BadSendOfHostInfo
      gErrStr = "Could not send SOCKS5 request packet"
    Case aeWinsockError_SOCKS5ReplyToSmall
      gErrStr = "SOCKS5 reply packet was of an incorrect length"
    Case aeWinsockError_SOCKS5VersioninReplyInvalid
      gErrStr = "SOCKS5 reply packet did not contain correct version"
    Case aeWinsockError_SOCKS5GeneralFail
      gErrStr = "SOCKS5 server failed"
    Case aeWinsockError_SOCKS5Rulesetfail
      gErrStr = "SOCKS5 reported connection was not allowed by ruleset"
    Case aeWinsockError_SOCKS5NetUnreach
      gErrStr = "SOCKS5 reported network was unreachable"
    Case aeWinsockError_SOCKS5HostUnreach
      gErrStr = "SOCKS5 reported host was unreachable"
    Case aeWinsockError_SOCKS5ConnectionRefused
      gErrStr = "SOCKS5 reported connection was refused"
    Case aeWinsockError_SOCKS5TTLExpired
      gErrStr = "SOCKS5 reported time to live expired"
    Case aeWinsockError_SOCKS5ComNoSupport
      gErrStr = "SOCKS5 reported command was not supported"
    Case aeWinsockError_SOCKS5AddrTypeNotSupported
      gErrStr = "SOCKS5 reported address type was not supported"
    Case aeWinsockError_SOCKS5UnAssignedError
      gErrStr = "SOCKS5 reported an error, but this error has not been assigned a reason"
    Case aeWinsockError_SOCKS5BadReplyNoRes
      gErrStr = "SOCKS5 replied with an invalid reply packet"
    Case aeWinsockError_SOCKSRepAddrTypeNoSupportByClient
      gErrStr = "SOCKS5 replied with hostname address this is not supported"
    Case aeWinsockError_SOCKSCloseLevel1
      gErrStr = "SOCKS5 closed the connection whilst requesting supported methods"
    Case aeWinsockError_SOCKSCloseLevel2
      gErrStr = "SOCKS5 closed the connection after sending supported method"
    Case aeWinsockError_SOCKSCloseLevel3
      gErrStr = "SOCKS5 closed the connection before username/password authentication could complete"
    Case aeWinsockError_SOCKSCloseLevel4
      gErrStr = "SOCKS5 closed the connection before a SOCKS request could be made correctly"
    Case aeWinsockError_SOCKSCloseLevel5
      gErrStr = "SOCKS5 closed the connection remote binded socket has been closed"
    Case Else
      gErrStr = "Error not defined"
  End Select
End Function

Private Function EvtFace(ByVal lPtr As Long) As aeSocketEvent
  Dim f, fT As aeSocketEvent
  '
  CopyMemory fT, lPtr, 4
  Set f = fT
  CopyMemory fT, 0&, 4
  f.thisaeSocket Me
  
  Set EvtFace = f
  '
End Function

Private Sub ApplySockOpt(h As sckOpt)
  With h
    KeepAlive = .bKeepAlive
    Linger = .bLinger
    OOBInLine = .bBandInLine
    ReUseAddress = .bReUseAddr
    DontRoute = .bDontRoute
    ReceiveBufferLength = .lGetBufLen
    SendBufferLength = .lSendBufLen
    ReceiveTimeout = .lGetTimeOut
    SendTimeout = .lSendTimeOut
    TcpNoDelay = .bTCPNoDelay
    Broadcast = .bAllowBroadcast
  End With
End Sub

Public Property Let LocalPort(ByVal vData As Long)
  On Error Resume Next
  
  With hData
    Select Case .hState
      Case Sck_Connected, Sck_Connecting, _
      Sck_Open, Sck_Listening
      Exit Property
    End Select
  End With
  
  If vData < 0 Then vData = 0
  With hData
    If .LocalPort <> vData Then
      .LocalPort = vData
    End If
  End With
End Property
Public Property Get LocalPort() As Long
    LocalPort = hData.LocalPort
End Property

Public Property Get LocalPeer() As String
    LocalPeer = hData.LocalPeer
End Property

Public Property Let RemotePort(ByVal vData As Long)
  On Error Resume Next
  
  With hData
    Select Case .hState
      Case Sck_Connected, Sck_Connecting, _
      Sck_Open, Sck_Listening
      Exit Property
    End Select
  End With

  If vData < 1 Then vData = 1
  With hData
    If .RemotePort <> vData Then
      .RemotePort = Int(vData)
    End If
  End With
End Property
Public Property Get RemotePort() As Long
    RemotePort = hData.RemotePort
End Property

Public Property Let RemotePeer(ByVal vData As String)
  On Error Resume Next
  With hData
    If .RemotePeer <> vData Then
      .RemotePeer = vData
      If IsHost(vData) = False Then
        .RemoteIP = vData
      Else
        .RemoteIP = ""
      End If
    End If
  End With
End Property
Public Property Get RemotePeer() As String
    RemotePeer = hData.RemotePeer
End Property

Public Property Get RemoteIP() As String
  RemoteIP = hData.RemoteIP
End Property

Public Property Let UseByteCount(ByVal vData As Boolean)
  If mvarUseByteCount <> vData Then
    mvarUseByteCount = vData
  End If
End Property
Public Property Get UseByteCount() As Boolean
    UseByteCount = mvarUseByteCount
End Property

Public Property Let UseEvents(ByVal vData As Boolean)
  If mvarUseEvents <> vData Then
    mvarUseEvents = vData
  End If
End Property
Public Property Get UseEvents() As Boolean
  If UseEvents <> mvarUseEvents Then
    UseEvents = mvarUseEvents
  End If
End Property

Public Property Let UseBuffer(ByVal vData As Boolean)
  If mvarUseBuffer <> vData Then
    mvarUseBuffer = vData
  End If
End Property
Public Property Get UseBuffer() As Boolean
    UseBuffer = mvarUseBuffer
End Property

Public Property Get UseSocketOptionInheritance() As Boolean
  UseSocketOptionInheritance = hData.bSocketReuse
End Property
Public Property Let UseSocketOptionInheritance( _
  ByVal vData As Boolean)
  With hData
    If .bSocketReuse <> vData Then .bSocketReuse = vData
  End With
End Property

Public Property Let ConnectionTimeout(ByVal vData As Long)
  On Error GoTo gErr:
  If vData <> mvarTimeout Then
    If vData < 0 Then vData = 1
    SetProp ws_hwnd, CStr(hData.hClassPtr & "#timercount"), 0
    mvarTimeout = vData
  End If
gErr:
  Err.Clear
End Property
Public Property Get ConnectionTimeout() As Long
    ConnectionTimeout = mvarTimeout
End Property

Public Property Get DNSTimeout() As Long
  DNSTimeout = cDNS.TimeOut
End Property
Public Property Let DNSTimeout(ByVal vData As Long)
  On Error GoTo gErr:
  cDNS.TimeOut = vData
gErr:
  Err.Clear
End Property

Public Property Get State() As aeSocketState
  State = hData.hState
End Property

Public Property Get StateDescription() As String
  Dim lhS As Long
  'Removed by SetState
  lhS = GetProp(ws_hwnd, CStr(hData.hClassPtr & "#ns"))
  'If it's not set, use global old state var
  If lhS = 0 Then lhS = hData.hState
  '
  Select Case lhS
    Case Sck_Closed
      StateDescription = "Closed"
    Case Sck_Open
      StateDescription = "Open"
    Case Sck_Listening
      StateDescription = "Listening"
    Case Sck_ConnectionPending
      StateDescription = "Connection Pending"
    Case Sck_ResolvingHost
      StateDescription = "Resolving Host"
    Case Sck_ResolvedHost
      StateDescription = "Resolved Host"
    Case Sck_Connecting
      StateDescription = "Connecting"
    Case Sck_Connected
      StateDescription = "Connected"
    Case Sck_Closing
      StateDescription = "Closing"
    Case Sck_Error
      StateDescription = "Error"
  End Select
End Property

Public Property Get Protocol() As aeSocketProtocol
  Protocol = hData.hProtocol
End Property
Public Property Let Protocol(ByVal vData As aeSocketProtocol)
  With hData
    If vData = .hProtocol Then Exit Property
    If .hSocket <> SOCKET_ERROR Then
      'Socket exists, must release current socket
      ReleaseSocket
      .hProtocol = vData
    Else
      'Socket doesn't exist, okay to just set the protocol
      .hProtocol = vData
    End If
  End With
End Property

Public Property Get IsWriteable() As Boolean
  IsWriteable = hData.sckWrite
End Property
Public Property Get IsReadable() As Boolean
  IsReadable = hData.sckRead
End Property

Public Property Get BytesSent() As Long
  BytesSent = hData.sntB
End Property
Public Property Get BytesReceived() As Long
  BytesReceived = hData.rcvB
End Property

Public Property Get SocketHandle() As Long
  SocketHandle = hData.hSocket
End Property

'// Socket Options

'The following code sets/gets socket options for TCP/UDP
'But as some options can't work with both protocols
'errors *are* expected, though the properties
'are stored because of the switching of protocol
'may result in these properties being set (and working)

'Also they are can be reused every time a new socket
'is created, so have to be kept, though they are not
'returned (in getting of options) only getsocketoption
'(getsockopt) is called if the socket exists

Public Property Get AcceptQueue() As Long
  AcceptQueue = hSck.lAcceptQ
End Property
Public Property Let AcceptQueue(ByVal vData As Long)
  If vData < 1 Then vData = 1
  With hSck
    If .lAcceptQ <> vData Then .lAcceptQ = vData
  End With
End Property

Public Property Get KeepAlive() As Boolean
  Dim APIVal As Long
  Dim lBool As Boolean
  With hData
    'Don't query option if no socket is created
    If .hSocket = SOCKET_ERROR Then Exit Property
    APIVal = getsocketoption(.hSocket, SOL_SOCKET, SO_KEEPALIVE, lBool, 4)
    'Didn't work
    If APIVal <> 0 Then Exit Property
    KeepAlive = lBool
  End With
End Property
Public Property Let KeepAlive(ByVal vData As Boolean)
  Dim lBool As Long
  If vData = True Then lBool = 1
  With hData
    With hSck
      If .bKeepAlive <> vData Then .bKeepAlive = vData
    End With
    If .hSocket = SOCKET_ERROR Then Exit Property
    setsocketoption .hSocket, SOL_SOCKET, SO_KEEPALIVE, lBool, 4
  End With
End Property

Public Property Get Broadcast() As Boolean
  Dim APIVal As Long
  Dim lBool As Boolean
  With hData
    'Don't query option if no socket is created
    If .hSocket = SOCKET_ERROR Then Exit Property
    APIVal = getsocketoption(.hSocket, SOL_SOCKET, SO_BROADCAST, lBool, 4)
    'Didn't work
    If APIVal <> 0 Then Exit Property
    Broadcast = lBool
  End With
End Property

Public Property Let Broadcast(ByVal vData As Boolean)
  Dim lBool As Long
  If vData = True Then lBool = 1
  With hData
    With hSck
     If .bAllowBroadcast <> vData Then .bAllowBroadcast = vData
    End With
    If .hSocket = SOCKET_ERROR Then Exit Property
    setsocketoption .hSocket, SOL_SOCKET, SO_BROADCAST, lBool, 4
  End With
End Property

Public Property Get Linger() As Boolean
  Dim strLO As LingerOptions
  getsocketoption hData.hSocket, SOL_SOCKET, SO_LINGER, strLO, 4
  Linger = strLO.l_OnOff <> 0
End Property
Public Property Let Linger(ByVal vData As Boolean)
  Dim strLO As LingerOptions
  With hSck
    If .bLinger <> vData Then .bLinger = vData
  End With
  If vData = True Then strLO.l_Linger = 1
  setsocketoption hData.hSocket, SOL_SOCKET, SO_LINGER, strLO, 4
End Property

Public Property Get OOBInLine() As Boolean
  Dim lBool As Boolean
  With hData
    'Socket doesn't exist!
    If .hSocket = SOCKET_ERROR Then Exit Property
    getsocketoption .hSocket, SOL_SOCKET, SO_OOBINLINE, lBool, 4
    OOBInLine = lBool
  End With
End Property
Public Property Let OOBInLine(ByVal vData As Boolean)
  Dim lBool As Long
  With hData
    With hSck
      'Save property for later use
      If .bBandInLine <> vData Then .bBandInLine = vData
    End With
    'Socket doesnt exist
    If .hSocket = SOCKET_ERROR Then Exit Property
    '
    lBool = vData
    'set socket
    setsocketoption .hSocket, SOL_SOCKET, SO_OOBINLINE, lBool, 4
  End With
End Property

Public Property Get ReUseAddress() As Boolean
  Dim lBool As Long
  With hData
    If .hSocket = SOCKET_ERROR Then Exit Property
    getsocketoption .hSocket, SOL_SOCKET, SO_REUSEADDR, lBool, 4
    ReUseAddress = lBool
  End With
End Property
Public Property Let ReUseAddress(ByVal vData As Boolean)
  Dim lBool As Long
  With hSck
    If .bReUseAddr <> vData Then .bReUseAddr = vData
  End With
  With hData
    If .hSocket = SOCKET_ERROR Then Exit Property
    lBool = vData
    setsocketoption .hSocket, SOL_SOCKET, SO_REUSEADDR, lBool, 4
  End With
End Property

Public Property Get DontRoute() As Boolean
  Dim lBool As Boolean
  With hData
    If .hSocket = SOCKET_ERROR Then Exit Property
    getsocketoption .hSocket, SOL_SOCKET, SO_DONTROUTE, lBool, 4
    DontRoute = lBool
  End With
End Property
Public Property Let DontRoute(ByVal vData As Boolean)
  Dim lBool As Long
  With hSck
    If .bDontRoute <> vData Then .bDontRoute = vData
  End With
  With hData
    If .hSocket = SOCKET_ERROR Then Exit Property
    lBool = vData
    setsocketoption .hSocket, SOL_SOCKET, SO_DONTROUTE, lBool, 4
  End With
End Property

Public Property Get ReceiveBufferLength() As Long
  With hData
    If .hSocket = SOCKET_ERROR Then Exit Property
    Dim lSize As Long
    getsocketoption .hSocket, SOL_SOCKET, SO_RCVBUF, lSize, 4
    ReceiveBufferLength = lSize
  End With
End Property
Public Property Let ReceiveBufferLength(ByVal vData As Long)
  With hSck
    If vData = 0 Then Exit Property
    If vData < 10 Then vData = 10
    If .lGetBufLen <> vData Then .lGetBufLen = vData
  End With
  With hData
    If .hSocket = SOCKET_ERROR Then Exit Property
    setsocketoption .hSocket, SOL_SOCKET, SO_RCVBUF, vData, 4
  End With
End Property

Public Property Get SendBufferLength() As Long
  With hData
    If .hSocket = SOCKET_ERROR Then Exit Property
    Dim lSize As Long
    getsocketoption .hSocket, SOL_SOCKET, SO_SNDBUF, lSize, 4
    SendBufferLength = lSize
  End With
End Property
Public Property Let SendBufferLength(ByVal vData As Long)
  With hSck
    If vData = 0 Then Exit Property
    If vData < 10 Then vData = 10
    If .lSendBufLen <> vData Then .lSendBufLen = vData
  End With
  With hData
    If .hSocket = SOCKET_ERROR Then Exit Property
    setsocketoption .hSocket, SOL_SOCKET, SO_SNDBUF, vData, 4
  End With
End Property

Public Property Get ReceiveTimeout() As Long
  With hData
    If .hSocket = SOCKET_ERROR Then Exit Property
    Dim lSize As Long
    getsocketoption .hSocket, SOL_SOCKET, SO_RCVTIMEO, lSize, 4
    ReceiveTimeout = lSize
  End With
End Property
Public Property Let ReceiveTimeout(ByVal vData As Long)
  With hSck
    If vData = 0 Then Exit Property
    If vData < 1 Then vData = 1
    If .lGetTimeOut <> vData Then .lGetTimeOut = vData
  End With
  With hData
    If .hSocket = SOCKET_ERROR Then Exit Property
    setsocketoption .hSocket, SOL_SOCKET, SO_RCVTIMEO, vData, 4
  End With
End Property

Public Property Get SendTimeout() As Long
  With hData
    If .hSocket = SOCKET_ERROR Then Exit Property
    Dim lSize As Long
    getsocketoption .hSocket, SOL_SOCKET, SO_SNDTIMEO, lSize, 4
    SendTimeout = lSize
  End With
End Property
Public Property Let SendTimeout(ByVal vData As Long)
  With hSck
    If vData = 0 Then Exit Property
    If vData < 1 Then vData = 1
    If .lSendTimeOut <> vData Then .lSendTimeOut = vData
  End With
  With hData
    If .hSocket = SOCKET_ERROR Then Exit Property
    setsocketoption .hSocket, SOL_SOCKET, SO_SNDTIMEO, vData, 4
  End With
End Property

Public Property Get TcpNoDelay() As Boolean
  Dim lBool As Boolean
  With hData
    If .hSocket = SOCKET_ERROR Then Exit Property
    getsocketoption .hSocket, SOL_SOCKET, TCP_NODELAY, lBool, 4
    TcpNoDelay = lBool
  End With
End Property

Public Property Let TcpNoDelay(ByVal vData As Boolean)
  Dim lBool As Long
  With hSck
    If .bTCPNoDelay <> vData Then .bTCPNoDelay = vData
  End With
  With hData
    If .hSocket = SOCKET_ERROR Then Exit Property
    lBool = vData
    setsocketoption .hSocket, SOL_SOCKET, TCP_NODELAY, lBool, 4
  End With
End Property

Private Sub aeDNSNotify_DNSNotify( _
  ByVal TaskHandle As Long, _
  ByVal HostPassed As String, _
  ByVal HostIP As String, _
  ByVal HostName As String, _
  ByVal HostAlias As String, _
  ByVal AbortedTask As Boolean, _
  ByVal Error As Boolean, _
  ByVal ErrorCode As Long, _
  ByVal ErrorCodeMessage As String, _
  ByVal TimeTaken As Long, _
  Optional ByVal hTag As Long = 0&)
  
  Select Case hTag
    'DNS results for SOCKS
      Case SOCKET_ERROR
        With hData
          'Unable to resolve proxy address
          If Error = True Then
              If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_NoProxyResolve, gErrStr(aeWinsockError_NoProxyResolve)
            Else
              With hS
                'Save IP address
                .ProxyIP = HostIP
                'Invoke open
                wsOpen .ThatIP, .ThatPort
              End With
          End If
        End With
        'Wait for wsOpen to return, exit
      Exit Sub
    'Port was invalid
      Case 0
        With hData
          If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_BadPort, gErrStr(aeWinsockError_BadPort)
        End With
        Exit Sub
    'DNS results for anything else
      Case Else
        With hData
          If Error = True Then
            'DNS Error
            If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error ErrorCode, ErrorCodeMessage
            Exit Sub
          Else
            'Success
            If .hEventPtr <> 0 Then EvtFace(.hEventPtr).DNSLookUpDone HostName, True, HostPassed, False, 0, ""
          End If
        End With
   
  End Select
  
  'Store passed hostname, it is valid
  hData.RemotePeer = HostPassed
  'Set the state
  SetState Sck_ResolvedHost
  'Can't call 'wsOpen' because this method will stay in scope (until wsOpen returns)
  'and wsOpen will call the DNS method in aeDNS again and that
  'will come back to this method
  'VB won't let that happen
  
  With hPack
    'Save IP in long form
    .host_in = inet_addr(HostIP)
    'save port
    .host_port = hTag
    'save pointer to this class
    .host_inv = hData.hClassPtr
  End With
  
  'Pass struct to timer proc which will execute 'wsOpen'
  'passing it the args, the hPack variable is private global
  'so no worrying about it going out of scope
  'This method will go out of scope after this line
  SetTimer ws_hwnd, VarPtr(hPack), 20, AddressOf wsOpenAsync

End Sub

Private Sub aeSocketEvent_Binded( _
  ByVal Port As Long, _
  ByVal bRemote As Boolean)
  If mvarUseEvents = False Then Exit Sub
  RaiseEvent Binded(Port, bRemote)
End Sub
Private Sub aeSocketEvent_DataRead( _
    ByVal BufferLen As Long, ByVal Buffered As aeSocketReadDataEnum, _
    ByVal DataPtr As Long, Data() As Byte, _
    ByVal IsOOB As Boolean)
    If mvarUseEvents = False Then Exit Sub
    RaiseEvent DataRead(BufferLen, Buffered, DataPtr, Data, IsOOB)
End Sub
Private Sub aeSocketEvent_Opened( _
    ByVal RemotePeer As String, ByVal RemotePort As Long, _
    ByVal LocalPeer As String, ByVal LocalPort As Long, _
    ByVal bFromListen As Boolean)
    If mvarUseEvents = False Then Exit Sub
    RaiseEvent Opened(RemotePeer, RemotePort, LocalPeer, LocalPort, bFromListen)
End Sub
Private Sub aeSocketEvent_AcceptConnection( _
    ByVal RemotePeer As String, ByVal RemotePort As Long, ByVal SocketID As Long)
    If mvarUseEvents = False Then Exit Sub
    RaiseEvent AcceptConnection(RemotePeer, RemotePort, SocketID)
End Sub
Private Sub aeSocketEvent_Closed( _
    ByVal RemotePeer As String, ByVal RemotePort As Long, _
    ByVal bRemoteClosure As Boolean, _
    ByVal bGrace As Boolean, ByVal bKeptData As Boolean)
    If mvarUseEvents = False Then Exit Sub
    RaiseEvent Closed(RemotePeer, RemotePort, bRemoteClosure, bGrace, bKeptData)
End Sub
Private Sub aeSocketEvent_DNSLookUpDone( _
    ByVal Host As String, ByVal bIsHost As Boolean, _
    ByVal PassedHost As String, _
    ByVal bError As Boolean, _
    ByVal ErrorCode As Long, ByVal ErrorCodeMessage As String)
    If mvarUseEvents = False Then Exit Sub
    RaiseEvent DNSLookUpDone(Host, bIsHost, PassedHost, bError, ErrorCode, ErrorCodeMessage)
End Sub
Private Sub aeSocketEvent_DNSLookUpStarted( _
    ByVal Host As String, ByVal bIsHost As Boolean)
    If mvarUseEvents = False Then Exit Sub
    RaiseEvent DNSLookUpStarted(Host, bIsHost)
End Sub
Private Sub aeSocketEvent_Error( _
    ByVal ErrorCode As Long, ByVal ErrorString As String)
    If mvarUseEvents = False Then Exit Sub
    RaiseEvent Error(ErrorCode, ErrorString)
End Sub
Private Sub aeSocketEvent_SocketChanged(ByVal NewHandle As Long, ByVal SelfPtr As Long)
    If mvarUseEvents = False Then Exit Sub
    RaiseEvent SocketChanged(NewHandle, SelfPtr)
End Sub
Private Sub aeSocketEvent_StateChanged(ByVal NewState As aeSocketState)
    If mvarUseEvents = False Then Exit Sub
    RaiseEvent StateChanged(NewState)
End Sub
Private Sub aeSocketEvent_thisaeSocket(ByVal ido As aeSocket)
  '
End Sub

Private Function aeSubclass_WindowProc(ByVal hwnd As Long, ByVal iMsg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long

  DoEvents

  Dim lErr As Long
  Dim lEvent As Long
 
  Select Case iMsg
    Case SOCKS_REPLY
      'Let timer code know not to post a timeout
      SetProp hwnd, CStr(hData.hClassPtr & "#ctimeout"), 1
      'Process errors before checking event
      lErr = GetAsyncError(lParam)
      Select Case lErr
        Case WSAETIMEDOUT
          lErr = aeWinsockError_ProxyConnTimeOut
        Case WSAECONNREFUSED
          lErr = aeWinsockError_NoConnectProxy
      End Select
      Select Case lErr
        Case 0
          'No Error
        Case aeWinsockError_NoConnectProxy, aeWinsockError_ProxyConnTimeOut, aeWinsockError_NegTimeOut
          'Can't use GetError here, as it can't use aeSocket error codes
          With hData
            If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error lErr, gErrStr(lErr)
          End With
          'Release current socket
          ReleaseSocket False, lErr
          Exit Function
        Case Else
          'Use winsock error messages
          GetError True, lErr
          ReleaseSocket (lEvent = FD_CLOSE), lErr
          Exit Function
      End Select
      'Get winsock event
      lEvent = GetSelectEvent(lParam)
      '
      Select Case lEvent
            
        Case FD_READ
          'Data on socket
          With hS
            If .TOC = SOCKS_V4 Then
              silSOCKS4Read wParam
            Else
              silSOCKS5Read wParam
            End If
          End With
          Exit Function
          
        Case FD_CONNECT
          'Set the state of the socket
          SetState SOCKS_Connected
            'Even if the socket connected we haven't yet got a proxied connection
            With hS
              If .TOC = SOCKS_V4 Then
                'send connect/bind request
                silSOCKS4Connect wParam
              Else
                'Request supported methods
                silSOCKS5Connect wParam
              End If
            End With
            
          Exit Function
          
        Case FD_CLOSE
          With hS
            If .TOC = SOCKS_V4 Then
              silSOCKS4Close wParam
            Else
              silSOCKS5Close wParam
            End If
          End With
          ReleaseSocket True
          Exit Function
          
        Case FD_OOB
        
      End Select
            
    Case SCK_REPLY
    
      'Stop timeout error
      DettachTimer
      'Process errors before checking event
      lErr = GetAsyncError(lParam)
      lEvent = GetSelectEvent(lParam)
      If lErr <> 0 Then
        'If we are in this "if" block
        'There is an error and geterror becomes passive.
        GetError True, lErr
        ReleaseSocket (lEvent = FD_CLOSE), lErr
        Exit Function
      End If
      
      Select Case lEvent
        Case FD_READ, FD_OOB
          silReadSocket wParam, lParam, lEvent = FD_OOB
          Exit Function
        
        Case FD_CONNECT, FD_CONNECTEX
          Dim APIVal As Long
          Dim scki As sockaddr_in
          Dim bLis As Boolean
          Dim bSOCKS As Boolean
          bSOCKS = hS.bUseSOCKS
          'If iMsg is FD_CONNECTEX, it's fake
          'therefore the connection is from a listening socket
          bLis = lEvent = FD_CONNECTEX
          'get remote information
          Select Case bSOCKS
          '
          Case True
          '
          With hData
            'Save client IP in remote host
            'Even though we are connected to the proxy
            .RemoteIP = hS.ThatIP
            'Save port information of client
            .RemotePort = hS.ThatPort
            'Return what SOCKS local port was used
            .LocalPort = hS.SeenPort
            'Return the IP the indirect connection see's us as
            .LocalPeer = hS.SeenIP
            'Mark the state of the socket as connected
            SetState Sck_Connected
            If .hEventPtr <> 0 Then
              'Raise connected event
              EvtFace(.hEventPtr).Opened .RemoteIP, .RemotePort, .LocalPeer, .LocalPort, bLis
            End If
          End With
          '
          Exit Function
          '
          Case False
          '
          APIVal = getpeername(wParam, scki, Len(scki))
          If APIVal = SOCKET_ERROR Then
            'failed (shouldn't really)
            GetError True
            ReleaseSocket
            Exit Function
          End If
          With hData
            'Get noted ip from ip long
            .RemoteIP = GetNotedIP(scki.sin_addr)
            'Get remote port information
            .RemotePort = htons(scki.sin_port)
          End With
          'Get local information
          APIVal = getsockname(wParam, scki, Len(scki))
          If APIVal = SOCKET_ERROR Then
            'Failed to get socket information
            GetError True
            ReleaseSocket
            Exit Function
          End If
          'Mark socket has connected
          SetState Sck_Connected
          With hData
            'Return what local port was used
            .LocalPort = htons(scki.sin_port)
            'get noted IP (this maybe the internet ip or not!)
            .LocalPeer = GetNotedIP(scki.sin_addr)
            If .hEventPtr <> 0 Then
              'Raise connected event
              EvtFace(.hEventPtr).Opened .RemoteIP, .RemotePort, .LocalPeer, .LocalPort, bLis
            End If
          End With
          '
          End Select
          '
          Exit Function
                
        Case FD_CLOSE
          ReleaseSocket True
          Exit Function
        
        Case FD_WRITE
          With hData
            If .sckWrite = False Then
              .sckWrite = True
            End If
          End With
        
        Case FD_ACCEPT
          silAccept wParam
          
      End Select
      
      Exit Function
     
    Case SOCKET_TIMER
      'If we've sent two timeouts turn timer off
      With hData
        If GetProp(hwnd, CStr(.hClassPtr & "#ctimeout")) <> 0 And GetProp(hwnd, CStr(.hClassPtr & "#ntimeout")) <> 0 Then
          GoTo TurnOff:
        End If
      End With
      If hS.NegTimeOut = 0 And mvarTimeout = 0 Then
TurnOff:
        'Turn off timer
        DettachTimer
        'Resetcount
        With hData
          RemoveProp hwnd, CStr(.hClassPtr & "#ctimeout")
          RemoveProp hwnd, CStr(.hClassPtr & "#ntimeout")
          RemoveProp hwnd, CStr(.hClassPtr & "#timercount")
          RemoveProp hwnd, CStr(.hClassPtr & "#negcount")
        End With
        Exit Function
      End If
      Dim lTC As Long
      Dim nTC As Long
      'Get the global count
      lTC = GetProp(hwnd, CStr(hData.hClassPtr & "#timercount"))
      'Get the negtimeout count
      nTC = GetProp(hwnd, CStr(hData.hClassPtr & "#negcount"))
      'Check count is greater or equal to negtime out
      If nTC >= hS.NegTimeOut Then
        With hData
          If GetProp(hwnd, CStr(.hClassPtr & "#ntimeout")) = 0 Then
          If .hState = SOCKS_Connected Then
            'Remove count
            RemoveProp hwnd, CStr(.hClassPtr & "#negcount")
            'flag we've timed out before
            SetProp hwnd, CStr(.hClassPtr & "#ntimeout"), 1
            'Only send time out in this case
              If hS.NegTimeOut <> 0 Then PostMessage hwnd, SOCKS_REPLY, .hSocket, ByVal WSAMakeSelectReply(FD_CLOSE, aeWinsockError_NegTimeOut)
            Exit Function
          End If
          End If
        End With
      Else
        'Inc the count
        nTC = nTC + 1
        SetProp hwnd, CStr(hData.hClassPtr & "#negcount"), nTC
        Exit Function
      End If
      'Check count is greater or equal to connection timeout
      If lTC >= mvarTimeout Then
          With hData
          If GetProp(hwnd, CStr(.hClassPtr & "#ctimeout")) = 0 Then
            'Remove count
            RemoveProp hwnd, CStr(.hClassPtr & "#timercount")
            'Flag we've sent a timeout
            SetProp hwnd, CStr(.hClassPtr & "#ctimeout"), 1
            'send timeout
            If mvarTimeout <> 0 Then
              If hS.bUseSOCKS = False Then
                'Send winsock timeout
                PostMessage hwnd, SCK_REPLY, .hSocket, ByVal WSAMakeSelectReply(FD_CLOSE, WSAETIMEDOUT)
              Else
                'send our timeout
                PostMessage hwnd, SOCKS_REPLY, .hSocket, ByVal WSAMakeSelectReply(FD_CLOSE, aeWinsockError_ProxyConnTimeOut)
              End If
            End If
          End If
          End With
        '
        Exit Function
        '
      End If
                
      'Inc the count
      lTC = lTC + 1
      'Save the count
      SetProp hwnd, CStr(hData.hClassPtr & "#timercount"), lTC

      Exit Function
  
  End Select
End Function

Private Sub Class_Initialize()
  'Set up properties
  With hData
    .hClassPtr = ObjPtr(Me)
    .hSocket = SOCKET_ERROR
    .hProtocol = sck_TCP
    SetState Sck_Closed
  End With
  With hS
    'Use V4
    .TOC = SOCKS_V4
    'Connect
    .Meth = SOCKSConnect
    '
    .NegTimeOut = 50
  End With
  'Events are raised by default
  mvarUseEvents = True
  'Set default event interface
  SetEventInterface Me
  'Create a DNS class
  Set cDNS = New aeDNS
  cDNS.TimeOut = 50
 
End Sub
Private Sub Class_Terminate()
  'ReleaseSocket information
  ReleaseSocket
  'Turn off timer
  DettachTimer
  'Cancel any current DNS look up
  cDNS.CancelDNS
  'Free DNS
  Set cDNS = Nothing
  
  'Free properties
  With hData
    .RemotePeer = ""
    .LocalPeer = ""
    .LocalPort = 0
    .hEventPtr = 0
    With .sck_in
      .sin_addr = 0
      .sin_family = 0
      .sin_port = 0
    End With
    'Remove properties
    RemoveProp ws_hwnd, CStr(.hClassPtr & "#ctimeout")
    RemoveProp ws_hwnd, CStr(.hClassPtr & "#ntimeout")
    RemoveProp ws_hwnd, CStr(.hClassPtr & "#timercount")
    RemoveProp ws_hwnd, CStr(.hClassPtr & "#negcount")
  End With
End Sub

Private Sub AttachTimer()
  'If already running, don't attach
  If mvarTimerRunning = True Then Exit Sub
  'If timeout is zero (disabled) don't attach
  If mvarTimeout = 0 And mvarTimerRunning = True Then
    DettachTimer
  Else
    If mvarTimerRunning = True Then Exit Sub
  End If
  'Reset count
  SetProp ws_hwnd, CStr(hData.hClassPtr & "#timercount"), 0
  'Attach Message
  HookTimer Me
  'set flag to remember subclass state
  mvarTimerRunning = True
End Sub
Private Sub DettachTimer()
  'If not running exit
  If mvarTimerRunning = False Then Exit Sub
  'Unattach message
  UnHookTimer Me
  'Set count to zero for connection time out
  SetProp ws_hwnd, CStr(hData.hClassPtr & "#timercount"), 0
  'Set neg count to zero
  SetProp ws_hwnd, CStr(hData.hClassPtr & "#negcount"), 0
  'Flag not runnning
  mvarTimerRunning = False
End Sub

Private Sub GetSocket()
  With hData
    If .hSocket <> SOCKET_ERROR Then
      'a socket already exists
      'close and free the handle
      ReleaseSocket
    End If
    
    Dim lSocket As Long
    Dim lErr As Long
    
    Select Case .hProtocol
      Case 1 'Create a TCP socket
        lSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)
        'Socket could not be made, report error and exit
        If lSocket = SOCKET_ERROR Then
          GetError True
          Exit Sub
        End If
        
      Case 2 'Create a UDP socket
        lSocket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)
        'Socket could not be made, report error and exit
        If lSocket = SOCKET_ERROR Then
          GetError True
          Exit Sub
        End If
     End Select
   
     If .hEventPtr <> 0 Then
        EvtFace(.hEventPtr).SocketChanged lSocket, .hClassPtr
     End If
     'store handle
     .hSocket = lSocket
    
    'Make sure this socket's message goto our subclass interface only
    SetInterface lSocket, Me, "SCK"
    
    'Tell Winsock to post messages for this socket :
    Dim APIVal As Long
    APIVal = WSAAsyncSelect(lSocket, ws_hwnd, SCK_REPLY, FD_ALL)
    If APIVal <> 0 Then
      'An error!
      GetError True
      ReleaseSocket False
      Exit Sub
    End If
    
    'Apply socket options
    If .bSocketReuse = True Then ApplySockOpt hSck
            
  End With
End Sub

'Clears information from 'hdata' like local peer/port
Private Sub ClearLocalInfo()
  With hData
    .LocalPort = 0
    .rcvB = 0
    .sntB = 0
    With .sck_in
      .sin_addr = 0
      .sin_family = 0
      .sin_port = 0
    End With
  End With
End Sub

Private Sub ReleaseSocket( _
  Optional bRemote As Boolean = False, _
  Optional lErr As Long = 0)
  
  With hData
    'No socket exists, exit
    If .hSocket = SOCKET_ERROR Then Exit Sub
    'Delete reference interface
    DeleteInterface .hSocket, "SCK"
    'Deltete SOCK interface
    DeleteInterface .hSocket, "SOCKS"
    '
    Dim oST As aeSocketState
    'Store old state
    oST = .hState
    'Stop Winsock posting any more messages
    WSAAsyncSelect .hSocket, ws_hwnd, 0, 0
    'Set state as tearing down socket
    SetState Sck_Closing
    'Stop timer
    DettachTimer
    'Stop DNS
    cDNS.CancelDNS
  
   'Make sure socket properties are false
    With hData
      .sckWrite = False
      .sckRead = False
    End With
    
   'shutdown socket for send/recv
    socket_shutdown .hSocket, 2
    'Close socket and release handle
    closesocket .hSocket
    'State socket is "closed"
    SetState Sck_Closed
    'Remove socket handle reference
    .hSocket = SOCKET_ERROR
        
   'raise closed
    If .hEventPtr <> 0 Then
      Select Case oST
        'Only raise closed when in open or connected or errord
      Case Sck_Connected, Sck_Open, Sck_Error
        EvtFace(.hEventPtr).Closed .RemoteIP, .RemotePort, bRemote, Not CBool(lErr), hBuf.Size > 0
      End Select
    End If
    
    'Clear properties
    ClearLocalInfo
    
    'Clear the buffer
    FlushBuffer
    
    With hS
      'Allow properties to have read/write access
      .Level = 0
      'Clear bind host
      .BindHost = ""
      'Clear the port
      .lBindPort = 0
      'Clear the IP the 3rd see's the SOCKS server as
      .SeenIP = ""
      'Clear port
      .SeenPort = 0
      'Clear the host that was connected to use via SOCKS
      .IndirectHost = ""
      'Clear the port it was connected from
      .IndirectPort = 0
    End With
    With hData
      'Allow connection timeouts to be sent
      SetProp ws_hwnd, CStr(.hSocket & "#ctimeout"), 0
      'Allow neg time outs to be sent
      SetProp ws_hwnd, CStr(.hSocket & "#ntimeout"), 0
      'Remove the connection timeout count
      RemoveProp ws_hwnd, CStr(.hClassPtr & "#timercount")
      'Remove the neg timeout count
      RemoveProp ws_hwnd, CStr(.hClassPtr & "#negcount")
    End With
  End With
End Sub

Public Sub SetEventInterface( _
  this As aeSocketEvent)
  'Only allow setting of new interface if there is no socket
  If hData.hSocket <> SOCKET_ERROR Then Exit Sub
  'All code will now point to the passed interface
  'this is an advanced feature and shouldn't be used
  'unless you know what you're doing!
  With hData
    If ObjPtr(this) = 0 Then
      'Was passed 'Nothing' requery to use default interface
      SetEventInterface Me
    Else
      'Save long pointer to interface
      .hEventPtr = ObjPtr(this)
    End If
  End With
End Sub

Private Sub SetState(ByVal ns As aeSocketState)
  'This entire class's only access to the state of the owning
  'socket is through this sub
  'If current state and new state are the same
  'no event is raised.
  With hData
    If ns = .hState Then Exit Sub
    If .hEventPtr <> 0 Then
      SetProp ws_hwnd, CStr(.hClassPtr & "#ns"), CLng(ns)
      EvtFace(.hEventPtr).StateChanged ns
      .hState = ns
      RemoveProp ws_hwnd, CStr(.hClassPtr & "#ns")
    End If
  End With
End Sub

Private Function GetError( _
    Optional bRaiseEvent As Boolean = True, _
    Optional wsErr As Long = 0) As Long
    
    On Error GoTo gErr:
            
    Dim lErr As Long
    'Get last error from winsock
    If wsErr = 0 Then
      lErr = WSAGetLastError
    Else
      'We've been passed an error code, we can't check
      'it's validity, assume it's from the callback
      lErr = wsErr
    End If
    Select Case lErr
      'There was no error or a would block error [ignore]
      Case 0, WSAEWOULDBLOCK
        Exit Function
      Case Else
        GetError = lErr
    End Select

    With hData
      'Make sure we have a pointer and are allowed to raise the event
      If .hEventPtr <> 0 And bRaiseEvent = True Then
        EvtFace(.hEventPtr).Error lErr, WSAGetLastErrorMessage(lErr)
      End If
    End With
    
    Exit Function
gErr:
    #If INDEBUG Then
      Debug.Print "aeSocket.GetError : " & Err.Description
    #End If
    Err.Clear
End Function

Private Sub TallySendOrGet( _
  Optional lSend As Long = 0, Optional lGet As Long = 0)
 
  With hData
    'Count sent bytes
    .sntB = .sntB + lSend
    If Err Then
      'On error (due to overflow?)
      .sntB = 0
      'clear error and sent bytes tally
      Err.Clear
      Exit Sub
    End If
    'Count recv bytes
    .rcvB = .rcvB + lGet
    If Err Then
      'On error (Due to overflow?)
      .rcvB = 0
      'Clear error and tally and exit
      Err.Clear
      Exit Sub
    End If
  End With
    
End Sub


'Returns:
'* 1 := success
'* 0 := error!

'* Note: This is a "connect" event, but as UDP never connects
'        And TCP does and both may need a DNS look up
'        They are merged
Public Function wsOpen( _
  Optional ByVal RemotePeerAddress As String, _
  Optional ByVal RemotePeerPort As Long) As Long
  
  'Trim off spaces
  RemotePeerAddress = Trim(RemotePeerAddress)
 
  'Can't use SOCKS over UDP
  With hS
    If hData.hProtocol = sck_UDP And .bUseSOCKS = True Then
      .bUseSOCKS = False
    End If
  End With
  
  With hData
  'Work out remote address, if we aren't passed one
  'We can use the stored one, if thats not present
  'error.
  If Len(RemotePeerAddress) = 0 Then
    If Len(.RemotePeer) = 0 Then
      If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_BadHost, gErrStr(aeWinsockError_BadHost)
      Exit Function
    Else
      RemotePeerAddress = .RemotePeer
    End If
  End If
  
  'Same as above, but as a long is used, assume 0 is invalid
  If RemotePeerPort = 0 Then
    If .RemotePort = 0 Then
      'bad port
      If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_BadPort, gErrStr(aeWinsockError_BadPort)
      Exit Function
    Else
      RemotePeerPort = .RemotePort
    End If
  Else
    If RemotePeerPort < 1 Then RemotePeerPort = 1
  End If
  
  End With
  
  If IsHost(RemotePeerAddress) = True Then
    'Resolve address
    With hData
      If .hEventPtr <> 0 Then
        EvtFace(.hEventPtr).DNSLookUpStarted RemotePeerAddress, True
      End If
    End With
    'close current socket
    ReleaseSocket
    '
    SetState Sck_ResolvingHost
    'DNS look up
    cDNS.DNS RemotePeerAddress, Me, False, RemotePeerPort
    Exit Function
    
  End If
  
  '
  ' RemotePeerAddress and RemotePeerPort now have a
  ' valid IP address/port
  '

  'Get a new one
  GetSocket
  'If socket couldn't be created exit
  'GetSocket fires error so we don't need to
  If hData.hSocket = SOCKET_ERROR Then Exit Function
 
  With hS
    Select Case .bUseSOCKS
      Case True 'Proxy'
        'Proxy hostname doesn't exist
        If Len(.ProxyName) = 0 Then
          .bUseSOCKS = False
          '
          GoTo NormalConnection
        End If
        'Port doesn't exist
        If .ProxyPort <= 0 Then
          .bUseSOCKS = False
          '
          GoTo NormalConnection
        End If
          'Store IP
          .ThatIP = RemotePeerAddress
          'Store Port
          .ThatPort = RemotePeerPort
        'Need to resolve proxy name!
        If Len(.ProxyIP) = 0 Then
          cDNS.DNS .ProxyName, Me, False, SOCKET_ERROR
          Exit Function
        End If
        
        'Can't make any more DNS calls, at this point
        'The aeDNSNotify is still in scope, its waiting
        'for this function to return
        
        With hS
          'Change access level (outside access to properties is read only)
          .Level = 1
          'Make the connection connect to the proxy ip
          RemotePeerAddress = .ProxyIP
          'Make the connection connect to the port
          RemotePeerPort = .ProxyPort
        End With
        With hData
          'Make winsock post SOCKS_REPLY instead.
          WSAAsyncSelect .hSocket, ws_hwnd, SOCKS_REPLY, FD_ALL
          'Delete old reference to socket|subclass interface
          DeleteInterface .hSocket, "SCK"
          'Set new SOCKS interface
          SetInterface .hSocket, Me, "SOCKS"
        End With
          '
    End Select
  End With
 
NormalConnection:
     
  With hData
    Select Case .hProtocol
      Case 1 'TCP
        'Fill sockaddr_in struct with host info
        With hData.sck_in
          .sin_addr = inet_addr(RemotePeerAddress)
          .sin_port = htons(RemotePeerPort)
          .sin_family = AF_INET
        End With
        'Bind a local port (if zero) skip this
        'winsock will bind a random port for us
        If .LocalPort <> 0 Then
          Dim sckB As sockaddr_in
          With sckB
            .sin_family = AF_INET
            .sin_addr = 0
            .sin_port = htons(.sin_port)
          End With
          'we don't care if it worked or not
          bind .hSocket, sckB, Len(sckB)
        End If
        'Mark socket as connecting (to host)
        'Errors are dealt with in the call back interface
        SetState Sck_Connecting
        'Set Timer
        AttachTimer
        'Connect to host
        Dim APIVal As Long
        APIVal = socketconnect(.hSocket, .sck_in, Len(.sck_in))
        
        If APIVal <> 0 Then
          'There was an error
          If GetError <> 0 Then
            'Looked up error, release socket
            ReleaseSocket
            Exit Function
          End If
        End If
        
      Case 2 'UDP
        'Save remote info, this data is needed to send when
        'using a UDP socket
        With hData.sck_in
          .sin_addr = inet_addr(RemotePeerAddress)
          .sin_port = htons(RemotePeerPort)
          .sin_family = AF_INET
        End With
        
        'Also save remote info so it can be accessed
        'externally
        .RemoteIP = RemotePeerAddress
        .RemotePort = RemotePeerPort
        
        'bind socket to a local port, this must be done
        'not needed for sending data (but for getting it!)
        'if zero is passed a random free port is used
        Dim scki As sockaddr_in
        With scki
          '.sin_addr = 0
          .sin_port = htons(hData.LocalPort)
          .sin_family = AF_INET
        End With
        APIVal = bind(.hSocket, scki, Len(scki))
        If APIVal = SOCKET_ERROR Then
          'Couldn't bind to passed port maybe in use
          GetError True
          ReleaseSocket
          Exit Function
        End If
        'need to assertain the local port information
        APIVal = getsockname(.hSocket, scki, Len(scki))
        If APIVal = SOCKET_ERROR Then
          GetError True
          ReleaseSocket
          Exit Function
        End If
   
        'Save port information
        .LocalPort = htons(scki.sin_port)
        
        'Mark socket as "open" ready to get/send data
        'As no connection can take place
        SetState Sck_Open
       
       'Raise fake "open" event,
        If .hEventPtr <> 0 Then
          EvtFace(.hEventPtr).Opened .RemotePeer, .RemotePort, .LocalPeer, .LocalPort, False
        End If

    End Select
  End With
  
  wsOpen = 1
 
End Function

Public Function wsClose() As Long
  'Store the amount of bytes the buffer DOES contain
  wsClose = hBuf.Size
  'Release socket, thereby closing it
  ReleaseSocket
  'Buffer has been released, return amount of bytes unread
End Function

'Returns:
' Nonzero := LocalPort that was binded (and is listening for TCP)
' Zero    := Didn't bind because of argument error
' -1      := Didn't bind because of winsock error

Public Function wsBindListen( _
  Optional LocalPeerPort As Long, _
  Optional IgnoreError As Boolean = False) As Long

  Dim lBPort As Long

  Select Case LocalPeerPort
    Case Is < 0
    'Make sure port number isn't less than zero
    If IgnoreError = True Then Exit Function
      With hData
        'Was less than zero, exit, error if allowed
        If .hEventPtr <> 0 Then
          EvtFace(.hEventPtr).Error aeWinsockError_BadPortArgumentMinus, gErrStr(aeWinsockError_BadPortArgumentMinus)
        End If
      End With
    Exit Function
    Case 0
      With hData
        'If localport (property) and LocalPeerPort(argument)
        'are both zero, zero is used,
        'if localport (property) is set, it's used
        If .LocalPort = 0 And LocalPeerPort = 0 Then
          lBPort = 0
        Else
          lBPort = .LocalPort
        End If
      End With
    Case Else
      lBPort = LocalPeerPort
  End Select
  
  'Call get socket, it will internally call
  'release socket if needed
  GetSocket
  'Socket wasn't created, error has been sent
  If hData.hSocket = SOCKET_ERROR Then
    wsBindListen = SOCKET_ERROR
    Exit Function
  End If
  
  'struct needed to be used with bind
  Dim sinB As sockaddr_in
  Dim APIVal As Long
  With sinB
    .sin_addr = 0
    .sin_family = AF_INET
    .sin_port = htons(lBPort)
  End With
  'bind socket to local port
  APIVal = bind(hData.hSocket, sinB, Len(sinB))
  If APIVal = SOCKET_ERROR Then
    GetError (Not IgnoreError)
    ReleaseSocket
    wsBindListen = SOCKET_ERROR
    Exit Function
  End If
  'Get the port that was binded
  Dim sckB As sockaddr_in
  APIVal = getsockname(hData.hSocket, sckB, Len(sckB))
  If APIVal = SOCKET_ERROR Then
    GetError (Not IgnoreError)
    ReleaseSocket
    wsBindListen = SOCKET_ERROR
    Exit Function
  End If
  
  With hData
    'Store what port was binded to
    .LocalPort = htons(sckB.sin_port)
    'Get what IP the port was binded to / may not work
    .LocalPeer = GetNotedIP(sckB.sin_addr)
    Select Case .hProtocol
      Case 1 'TCP
        'Start a listen!
        APIVal = listen(.hSocket, hSck.lAcceptQ)
        If APIVal = SOCKET_ERROR Then
          'Couldn't start a listen, error/exit
          GetError (Not IgnoreError)
          ReleaseSocket
          wsBindListen = SOCKET_ERROR
          Exit Function
        End If
    End Select
  End With
  
  'We can't use the "open" state for UDP as
  'wsSend/Ex need a remote IP for UDP, and we don't have
  'that yet (until a remote machine sends us data)
  
  'The winsock control marks the socket "open" but
  'won't let you send data in any way.
  'Even if you set a remote IP.
  
  SetState Sck_Listening
  'Return Port, raise bind event
  With hData
    wsBindListen = .LocalPort
    If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Binded .LocalPort, False
  End With
End Function

'Return:
'* nonzero := Amount of data actually sent.
'
'* Sends: Byte Array, Long Array, Integer Array
'         Boolean Array, String Array
'         String, Long, Integer, Byte, Any UDT (with fixed length strings)
'         across a TCP/UDP socket
'
'  Can send : Normally, Out of band, No Route)
Public Function wsSend( _
  this As Variant, _
  Optional SendFlags As aeSocketSendOptions = SendNormal, _
  Optional UserDefinedTypeVarPtr As Long = 0, _
  Optional UserDefinedTypeLen As Long = 0, _
  Optional IgnoreError As Boolean = True) As Long
  
  On Error GoTo gErr:
    
  'Check if we have a state ready to send,
  Select Case hData.hState
    Case Sck_Open, Sck_Connected
      'Socket is either (open) UDP or (connected) TCP
    Case Else
      'We can't send, socket isn't in the right state here
      With hData
        'Ignore error if requested.
        If IgnoreError = True Then Exit Function
        'if interface exists report error and exit
        If .hEventPtr <> 0 Then
          EvtFace(.hEventPtr).Error aeWinsockError_BadState2Send, gErrStr(aeWinsockError_BadState2Send)
        End If
      End With
      Exit Function
  End Select
   
  'Turn the variant data into a byte array
  Dim b() As Byte
  Dim bSize As Long
  Dim bFlag As Long
  
  If (UserDefinedTypeVarPtr > 0) And (UserDefinedTypeLen > 0) Then
  
      'WARNING: Do NOT pass any old number for either of these
      'parameters, doing so will crash the code!
      'You should not pass a variant array
      'I only provide this feature for advanced users
      
      'NOTE: You *can* send an array of a userdefined type
      '      Make sure you pass the variable pointer to
      '      the FIRST index, and the length is the size of
      '      the entire array times the size of one index
      ' eg
      '      wsSend Nothing, , VarPtr(UDT(0)), ((UBound(UDT) - LBound(UDT) + 1) * Len(UDT(0)))
      
      'Use of UDTs with strings that are not of fixed length
      'has not been tested
      
      'Ignore passed 'this' variant and convert
      'from pointer the UDT to byte array
      
      'The user must correctly pass the length
      'of the userdefined type!
      ReDim b(0 To UserDefinedTypeLen - 1) As Byte
      
      'Copy from pointer to the UDT to the first index
      'of our byte array
      CopyMemory ByVal VarPtr(b(0)), ByVal UserDefinedTypeVarPtr, UserDefinedTypeLen
      
      'Save size
      bSize = (UBound(b) - LBound(b) + 1)
    
      'Don't process 'this'
      GoTo SkipConversion
  End If
  
  'Check if this is an array
  'and if so if the array is actually holding anything
  
  If IsArray(this) = True Then
     If IsArrayEmpty(this) = True Then
      'Can't deal with empty array
      With hData
        If .hEventPtr <> 0 Then
          EvtFace(.hEventPtr).Error aeWinsockError_BadArray, gErrStr(aeWinsockError_BadArray)
        End If
      End With
      Exit Function
    End If
    Dim lbHigh As Long
    Dim lbLow As Long
    'Save first and last index
    lbHigh = UBound(this)
    lbLow = LBound(this)
  End If
  
  Select Case VarType(this)
  
    Case vbString
      If Len(this) = 0 Then
        'We can't send an empty string!
        If IgnoreError = True Then Exit Function
        With hData
          If .hEventPtr <> 0 Then
            EvtFace(.hEventPtr).Error aeWinsockError_CantSendEmptyString, gErrStr(aeWinsockError_CantSendEmptyString)
          End If
        End With
        Exit Function
      End If
      'Convert variant string into a byte array
      b = StrConv(this, vbFromUnicode)
      
   Case vbArray Or vbByte
      Dim thisB() As Byte
      'Copy from the variant byte array to a byte array
      thisB = this
      'Resize array to match the this() array
      'but change the starting index
      ReDim b(0 To (lbHigh - lbLow)) As Byte
      'Copy from this() to b()
      CopyMemory ByVal VarPtr(b(0)), thisB(lbLow), lbHigh
       
   Case vbArray Or vbString
      'Winsock control does not support the sending
      'of this data type
      '(With good reason!)
      'Turn array of strings into one string, then into
      'a byte array
      'if anyone can come up with a better method?!
      Dim f As Long
      Dim thisStr As String
      'Go through each index and copy to a string type
      For f = lbLow To lbHigh
        thisStr = thisStr & this(f)
      Next
      'Convert to byte array from string
      b = thisStr
      
   Case vbArray Or vbInteger
      Dim bInt() As Integer
      'Create a temp integer array
      bInt = this
      'Size the byte array to fit integer array
      ReDim b(0 To ((lbHigh - lbLow) * 2) + 1) As Byte
      
      'Copy from integer array to byte array
      CopyMemory ByVal VarPtr(b(0)), ByVal VarPtr(bInt(lbLow)), (UBound(b) - LBound(b) + 1)
      
      '// Code to convert back into a integer array
      'Dim t() As Integer
      'ReDim t(0 To (UBound(b) \ 2)) As Integer
      'CopyMemory ByVal VarPtr(t(0)), ByVal VarPtr(b(0)), (UBound(b) - LBound(b) + 1)
      '\\
      
    Case vbArray Or vbLong
      Dim bLong() As Long
      'Copy variant long array into a defined long array
      bLong = this
      'Resize so byte array is exactly the same in bytes
      'as the long array
      ReDim b(0 To ((lbHigh - lbLow) * 4) + 1) As Byte
      
      'Copy from long array to byte array
      CopyMemory ByVal VarPtr(b(0)), ByVal VarPtr(bLong(lbLow)), (UBound(b) - LBound(b) + 1)
      
      '// Code to convert back to long array
      'Dim t() As Long
      'ReDim t(0 To (UBound(b) \ 4)) As Long
      'CopyMemory ByVal VarPtr(t(0)), ByVal VarPtr(b(0)), (UBound(b) - LBound(b) + 1)
      '\\
    Case vbArray Or vbBoolean
      Dim bBool() As Boolean
      'Copy the variant boolean array to a defined
      'boolean array
      bBool = this
      'Resize array
      ReDim b(0 To ((lbHigh - lbLow) * 2) + 1) As Byte
      'Copy from boolean array to long array
      CopyMemory ByVal VarPtr(b(0)), ByVal VarPtr(bBool(lbLow)), (UBound(b) - LBound(b) + 1)
      
      '// Code to convert back to a boolean array
      'Dim t() As Boolean
      'ReDim t(0 To (UBound(b) \ 2)) As Boolean
      'CopyMemory ByVal VarPtr(t(0)), ByVal VarPtr(b(0)), (UBound(b) - LBound(b) + 1)
      '\\
    Case vbLong
      'Resize byte array to fit long (4bytes)
      ReDim b(0 To 3) As Byte
      'Copy from long to byte array
      CopyMemory ByVal VarPtr(b(0)), CLng(this), 4
      
      '// Code to convert from byte array to long
      'Dim lTmp As Long
      'CopyMemory ByVal VarPtr(lTmp), ByVal VarPtr(b(0)), 4
      '\\
      
    Case vbInteger
      ReDim b(0 To 1) As Byte
      CopyMemory ByVal VarPtr(b(0)), CInt(this), 2
      
      '// Code to convert from a byte array to an int
      'Dim lInt As Integer
      'CopyMemory lInt, ByVal VarPtr(b(0)), 2
      '\\
      
    Case Else
   
      'IF in this case block, we can't use the type
      'inside the variant,
      
      'don't report error
      If IgnoreError = True Then Exit Function
        With hData
          If .hEventPtr <> 0 Then
            EvtFace(.hEventPtr).Error aeWinsockError_CantUseDataType, gErrStr(aeWinsockError_CantUseDataType)
          End If
        End With
      Exit Function
  End Select
   
  'Find the size of b()
  bSize = (UBound(b) - LBound(b) + 1)
    
SkipConversion:
 
  'Work out flags, enum and winsock defined flags
  'don't match anymore
  If (SendFlags And SendNormal) = SendNormal Then
    bFlag = MSG_NORMAL
  End If
  If (SendFlags And SendNoRoute) = SendNoRoute Then
    bFlag = bFlag Or MSG_DONTROUTE
  End If
  If (SendFlags And SendOOB) = SendOOB Then
    bFlag = bFlag Or MSG_OOB
  End If
 
  Dim APIVal As Long
  With hData
    Select Case .hProtocol
      Case 1 'TCP
        'Send byte array through connected socket
        APIVal = send(.hSocket, b(0), bSize, bFlag)
      Case 2 'UDP
        'Send byte array through unconnected socket
        'the sock struct has been filled in the 'wsOpen' event
        APIVal = sendto(.hSocket, b(0), bSize, bFlag, .sck_in, Len(.sck_in))
    End Select
    
    If APIVal = SOCKET_ERROR Then
      'We didn't send the data
      If GetError(True) <> 0 Then
        'There was an error
        ReleaseSocket False
      Else
        'We got a would block message!
        .sckWrite = False
      End If
      wsSend = SOCKET_ERROR
      Exit Function
    End If
    
    'count bytes?
    If UseByteCount = True Then TallySendOrGet APIVal
    
    'return amount actually sent in bytes
    wsSend = APIVal
  End With
  
Exit Function
gErr:
  With hData
    If .hEventPtr <> 0 Then
      'Raise error, and ignore if 'ignore error' true
      EvtFace(.hEventPtr).Error aeWinsockError_FailSendInternal, gErrStr(aeWinsockError_FailSendInternal)
    End If
  End With
  Err.Clear
End Function

'* Send any memory object thru a TCP/UDP socket
'* Same as wsSend in all other regards.
Public Function wsSendEx( _
  hMem As Long, _
  Optional SendFlags As aeSocketSendOptions = SendNormal, _
  Optional IgnoreError As Boolean = False) As Long
  
  '
  ' Note, When sending memory objects remember
  '       That the actual allocated size maybe greater
  '       Than requested size made to GlobalAlloc API
  '
  On Error GoTo gErr:
    
  'Check if we have a state ready to send,
  Select Case hData.hState
    Case Sck_Open, Sck_Connected
      'Socket is either (open) UDP or (connected) TCP
    Case Else
      'We can't send, socket isn't in the right state here
      With hData
        'Ignore error if requested.
        If IgnoreError = True Then Exit Function
        'if interface exists report error and exit
        If .hEventPtr <> 0 Then
          EvtFace(.hEventPtr).Error aeWinsockError_BadState2Send, gErrStr(aeWinsockError_BadState2Send)
        End If
      End With
      Exit Function
  End Select
  
  If hMem <= 0 Then
    'Ignore error, if requested
    If IgnoreError = True Then Exit Function
    With hData
      If .hEventPtr <> 0 Then
        'Raise error if event interface exists
        EvtFace(.hEventPtr).Error aeWinsockError_BadMemory, gErrStr(aeWinsockError_BadMemory)
      End If
    End With
    Exit Function
  End If
  
  Dim lMemSize As Long
  Dim bFlag As Long
  Dim lPtr As Long
  'Use Global size to see if the memory block is legal.
  lMemSize = GlobalSize(hMem)
  If lMemSize = 0 Then
    'Memory points to nothing we can use!.
    If IgnoreError = True Then Exit Function
    With hData
      If .hEventPtr <> 0 Then
        EvtFace(.hEventPtr).Error aeWinsockError_NoMemory, gErrStr(aeWinsockError_NoMemory)
      End If
    End With
    Exit Function
  End If
  
  'Lock down memory and get a pointer to the first byte
  lPtr = GlobalLock(hMem)
  If lPtr = 0 Then
    'Ignore error if requested
    If IgnoreError = True Then Exit Function
    With hData
      If .hEventPtr <> 0 Then
        EvtFace(.hEventPtr).Error aeWinsockError_NoPtr, gErrStr(aeWinsockError_NoPtr)
      End If
    End With
    Exit Function
  End If

  'Work out flags, enum and winsock defined flags
  'don't match anymore
  If (SendFlags And SendNormal) = SendNormal Then
    bFlag = MSG_NORMAL
  End If
  If (SendFlags And SendNoRoute) = SendNoRoute Then
    bFlag = bFlag Or MSG_DONTROUTE
  End If
  If (SendFlags And SendOOB) = SendOOB Then
    bFlag = bFlag Or MSG_OOB
  End If
  
  Dim APIVal As Long
  With hData
    Select Case .hProtocol
      Case 1 'TCP
        'Send pointer to memory thru socket
        APIVal = send(.hSocket, ByVal lPtr, lMemSize, bFlag)
      Case 2 'UDP
        'Send pointer to memory thru unconnected socket
        APIVal = sendto(.hSocket, ByVal lPtr, lMemSize, bFlag, .sck_in, Len(.sck_in))
    End Select
    
    'Unlock memory object, free the pointer
    GlobalUnlock hMem
    'Null pointer
    lPtr = 0
    
    If APIVal = SOCKET_ERROR Then
      'We didn't send the data
      If GetError(True) <> 0 Then
        'There was an error
        ReleaseSocket False
      Else
        'We got a would block message!
        .sckWrite = False
      End If
      wsSendEx = SOCKET_ERROR
      Exit Function
    End If
    
    'count bytes?
    If UseByteCount = True Then TallySendOrGet APIVal
    
    'return amount actually sent in bytes
    wsSendEx = APIVal
  End With
 
Exit Function
gErr:
  With hData
    If .hEventPtr <> 0 Then
      'Raise error, and ignore if 'ignore error' true
      EvtFace(.hEventPtr).Error aeWinsockError_FailSendInternal, gErrStr(aeWinsockError_FailSendInternal)
    End If
  End With
  Err.Clear
End Function

'Returns
'
' * 0 := Error!
' * 1 := Success
'
Public Function wsAccept( _
  ByVal SocketID As Long, _
  Optional IgnoreError As Boolean = False) As Long
  
  'The socketid is looked up in a stored table
  'this means any connection from a bind/listen accept pair
  'made by any aeSocket class can be accepted and is valid
  
  'But only new connections, you can never do :
  ' a.wsAccept b.SocketHandle
  
  'winsock control requestids can not be passed
  
  If GetProp(ws_hwnd, "socketbaby#" & SocketID) <> 1 Then
    'This socket was not accepted by a aeSocket class
    'or isn't valid, blah
    If IgnoreError = True Then Exit Function
    With hData
      If .hEventPtr <> 0 Then
        'Accept ID wasn't valid
        EvtFace(.hEventPtr).Error aeWinsockError_AcceptSocketInvalid, gErrStr(aeWinsockError_AcceptSocketInvalid)
      End If
    End With
    Exit Function
  Else
    'Remove property, it's job is done
    RemoveProp ws_hwnd, "socketbaby#" & SocketID
  End If
  
  'Release current socket if any exists
  ReleaseSocket
  With hData
    'Set protocol
    .hProtocol = sck_TCP
    'Set socket handle returned from accept()
    .hSocket = SocketID
    'Reapply socket options
    If .bSocketReuse = True Then ApplySockOpt hSck
    'Don't need to reapply event mask (WSASelect)
    'As the new socket takes the event mask of the listening
    'socket (FD_ALL)
    '
    'Reference socket to subclass interface
    SetInterface SocketID, Me, "SCK"
  End With
  
  'Winsock won't post a FD_CONNECT, the connection is made
  'with accept()
  'posting a fake FD_CONNECTEX will insure we read the peer info
  'it will let the connect code know to set 'bFromListen' to true
  SendMessage ws_hwnd, SCK_REPLY, SocketID, ByVal WSAMakeSelectReply(FD_CONNECTEX, 0)
  
  'Set property so 'silAccept' doesn't close the pending
  'connection.
  SetProp ws_hwnd, "SOCKETACCEPT#" & SocketID, 1
  
  'Return success
  wsAccept = 1
End Function

'Returns:
' True  := Data *was* read into [this]
' False := Data wasn't
'
'Inputs:
' this        := Variant type where the data will be copied
' WantDataLen := Amount in bytes to copy, zero to copy all
' bPeekData   := if True data won't be removed from the buffer
'                just copied
Public Function wsGet( _
  this As Variant, _
  Optional ByVal WantDataLen As Long = 0, _
  Optional bPeekData As Boolean = False) As Boolean
  
  On Error GoTo gErr:
  
  Dim b() As Byte
  Dim lPtr As Long
  Dim lPtr2 As Long
  Dim bProcess As Boolean
  Dim lMem As Long
  
  With hBuf
    'No data in buffer exit!
    If .Size = 0 Or .hMem = 0 Then Exit Function
    Select Case WantDataLen
      'Passed length is -1 or zero, give back all data
      Case Is <= 0
        WantDataLen = .Size
      'Requested size is greater than actual buffer readjust
      Case Is > hBuf.Size
        WantDataLen = .Size
    End Select
    
    'Make sure nothing is writing while we are 'adjusting'
    If .DontWrite = False Then
      .DontWrite = True
      'Get lock pointer
      lPtr = GlobalLock(.hMem)
      If lPtr Then
        'Adjust byte array
        ReDim b(0 To WantDataLen - 1) As Byte
        'Copy from pointer to b()
        CopyMemory b(0), ByVal lPtr, WantDataLen
        'Unlock memory object
         GlobalUnlock .hMem
        'Data has to be removed
        If bPeekData = False Then
          If .Size = WantDataLen Then
            'The entire block was read, just delete it
            GlobalFree .hMem
            'set the handle
            .hMem = 0
            'No size
            .Size = 0
            '
            .ActualTotalSize = 0
            'Reset flag
            .DontWrite = False
            GoTo ProcessVariant
          End If
          'Allocate new memory block
          lMem = GlobalAlloc(GMEM_MOVEABLE, .Size - WantDataLen)
          'get first lock pointer
          lPtr = GlobalLock(.hMem)
          'get lock pointer
          lPtr2 = GlobalLock(lMem)
          If lPtr And lPtr2 Then
            'Copy from old memory to new block
            CopyMemory ByVal lPtr2, ByVal lPtr + WantDataLen, (.Size - WantDataLen)
            'Unlock old memory
            GlobalUnlock .hMem
            'Unlock new memory
            GlobalUnlock lMem
            'Adjust the size
            .Size = .Size - WantDataLen
            'Change the write point
            .WritePoint = .WritePoint - WantDataLen
            'Free old memory
            GlobalFree .hMem
            'Save the new handle
            .hMem = lMem
            'save the size
            .ActualTotalSize = GlobalSize(.hMem)
          End If
        End If
        'b() has valid data
        bProcess = True
      End If
    End If
    'Allow writing to happen
    .DontWrite = False
  End With
  
  If bProcess = False Then Exit Function
ProcessVariant:
  
  'Find which data type 'this' is
  Select Case VarType(this)
  
    Case vbString, vbEmpty
      'Convert byte array to UNICODE string
      this = StrConv(b, vbUnicode)
      'Return success
      wsGet = True
 
    Case vbArray Or vbByte
      'this() is a byte array but is passed as a variant
      this = b
      'Return success
      wsGet = True
    Case Else
      'Can't support type
  End Select
  
Exit Function
gErr:
  Debug.Print "wsGet ERR: " & Err.Description
  Err.Clear
End Function

Private Sub silReadSocket(ByVal wParam As Long, _
  ByVal lParam As Long, IsOOB As Boolean)
  
  'wParam := Socket Handle
  'lParam := Error code, etc.
  'OOB    := IsData out of band?
  
  Dim bFlag As Long
  Dim APIVal As Long
  Dim lSize As Long
  Dim DRev As sockaddr_in
  Dim b() As Byte
  Dim lPtr As Long
  Dim hMem2 As Long
    
    'Get the amount of data if any
    ioctlsocket wParam, FIONREAD, lSize
  
    'No data to read!
    If lSize = 0 Then Exit Sub
 
    If IsOOB = True Then
      bFlag = MSG_NORMAL Or MSG_OOB
      Else
      bFlag = MSG_NORMAL
    End If
       
      'Resize array
      ReDim b(0 To lSize - 1) As Byte

      Select Case hData.hProtocol
         Case 1 'TCP
           APIVal = recv(wParam, b(0), lSize, bFlag)
         Case 2 'UDP
           APIVal = recvfrom(wParam, b(0), lSize, bFlag, DRev, Len(DRev))
           'Check address
           silGetUDPAddr DRev
      End Select
     'Check Error
     If APIVal = SOCKET_ERROR Then
       'didnt get data
       If GetError(True) <> 0 Then
         'There was an error
         ReleaseSocket False
       Else
         'We got a would block message!
         hData.sckRead = False
       End If
       Exit Sub
     End If
     'count bytes
     If UseByteCount = True Then TallySendOrGet , APIVal
     With hData
       If .hEventPtr <> 0 Then
         Select Case mvarUseBuffer
          Case False 'Don't buffer data
            EvtFace(.hEventPtr).DataRead APIVal, DataDirectStream, VarPtr(b(0)), b, IsOOB
          Case True 'Buffer data
            With hBuf
              If .hMem <> 0 Then
                'Memory object DOES exist
                  If .DontWrite = False Then
                    'Flag we are writing
                    .IsWriting = True
                    '
                    hMem2 = GlobalReAlloc(.hMem, .Size + lSize, GMEM_MOVEABLE)
                    If hMem2 Then
                      .ActualTotalSize = GlobalSize(.hMem)
                      'Lock/Get pointer
                      lPtr = GlobalLock(.hMem)
                      If lPtr Then
                        If .WritePoint <> SOCKET_ERROR Then
                          'Copy to pointer from b()
                          CopyMemory ByVal lPtr + .WritePoint, b(0), lSize
                          'save the new size
                          .Size = .Size + lSize
                          'Save the pointr index to write to
                          .WritePoint = .WritePoint + lSize
                        End If
                        'unlock/release pointer
                        GlobalUnlock .hMem
                      End If
                      'Flag we aren't writing
                      .IsWriting = False
                    End If
                  End If
              Else
                'Memory object does not exist
                If .DontWrite = False Then
                  'Alloc the memory
                  .hMem = GlobalAlloc(GMEM_MOVEABLE, lSize)
                  'Save the size
                  .ActualTotalSize = GlobalSize(.hMem)
                  'Flag we are writing
                  .IsWriting = True
                  'Inc memory lock count
                  lPtr = GlobalLock(.hMem)
                  If lPtr Then
                    'Copy to pointer from b() array
                    CopyMemory ByVal lPtr, b(0), lSize
                    'Unlock memory
                    GlobalUnlock .hMem
                  End If
                  'save the size
                  .Size = lSize
                  'flag we aren't writing
                  .IsWriting = False
                  'Where to start writing to
                  .WritePoint = lSize
                End If
              End If
            End With
            EvtFace(.hEventPtr).DataRead APIVal, DataBufferStream, VarPtr(b(0)), b, IsOOB
         End Select
       End If
     End With
      
End Sub

Private Sub silGetUDPAddr(idp As sockaddr_in)
  With hData
    If .sck_in.sin_addr = 0 Then
      'store struct
      .sck_in = idp
      'Save IP
      .RemoteIP = GetNotedIP(idp.sin_addr)
      'Save port
      .RemotePort = htons(idp.sin_port)
      'set the state of the socket
      SetState Sck_Open
      'Raise fake open event
      If .hEventPtr <> 0 Then
        With hData
          EvtFace(.hEventPtr).Opened .RemoteIP, .RemotePort, .LocalPeer, .LocalPort, True
        End With
      End If
      Exit Sub
    End If
  End With
End Sub

Private Sub silAccept(ByVal wParam As Long)
  
  Dim sckA As sockaddr_in
  Dim lSck As Long
  
  'Accept connection internally
  'Only way to get remote host info!
  lSck = accept(wParam, sckA, Len(sckA))
  If lSck = SOCKET_ERROR Then
    'Couldn't accept connection,
    GetError True
    ReleaseSocket True
    Exit Sub
  End If
  'property removed by 'wsAccept'
  SetProp ws_hwnd, "socketbaby#" & lSck, 1
  With hData
    If .hEventPtr <> 0 Then
      EvtFace(.hEventPtr).AcceptConnection GetNotedIP(sckA.sin_addr), htons(sckA.sin_port), lSck
    End If
    If GetProp(ws_hwnd, "SOCKETACCEPT#" & lSck) = 0 Then
      'Connection wasn't accepted close the socket
      closesocket lSck
    Else
      'Remove property
      RemoveProp ws_hwnd, "SOCKETACCEPT#" & lSck
    End If
  End With
  
End Sub

Private Sub silLevel(ByVal NewLevel As Long)
  With hS
    If NewLevel <> .Level Then
      .Level = NewLevel
    End If
  End With
End Sub

Private Sub silSOCKS5Connect(ByVal wParam As Long)
  'V5
  Dim APIVal As Long
  Dim b() As Byte
  Dim bSize As Long
  Dim S5 As String
  
  With hS
    'Don't carry on if in wrong level
    If .Level >= 2 Then Exit Sub
    'Change level
    silLevel 2
    'Build supported method request packet
    '-------------------
    ' 5 <version>
    ' 1 <number of methods>
    ' 0 <method> ...
    '-------------------
    'Use SOCKS5 protocol
    S5 = Chr(5)
    'methods :
     '---------------------
'    o x '00' NO AUTHENTICATION REQUIRED *
'    o x '01' GSSAPI
'    o x '02' USERNAME/PASSWORD *
'    o x '03' to X'7F' IANA ASSIGNED
'    o x '80' to X'FE' RESERVED FOR PRIVATE METHODS
'    May return:
'    o x 'FF' NO ACCEPTABLE METHODS
    '---------------------
    '* What we support
    If .bUP = True Then
      'Both username/password have been passed
      S5 = S5 & Chr(2) & Chr(0) & Chr(2)
    Else
      'No username/password request no auth only
      S5 = S5 & Chr(1) & Chr(0)
    End If
    'Resize b
    ReDim b(0 To Len(S5) - 1) As Byte
    'Convert the UNICODE string into a byte array
    b = StrConv(S5, vbFromUnicode)
    'Get/save the size of b()
    bSize = (UBound(b) - LBound(b) + 1)
    'Send b() to server
    APIVal = send(wParam, b(0), bSize, MSG_NORMAL)
    '
    If APIVal = SOCKET_ERROR Then
      If GetError(True) <> 0 Then
        'Couldn't send release
        ReleaseSocket False
      End If
      Exit Sub
    End If
    'Check size sent
    If APIVal <> bSize Then
      With hData
        'Error
        If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_CantSendMethodReqV5, gErrStr(aeWinsockError_CantSendMethodReqV5)
      End With
      ReleaseSocket False
    End If
  End With
  
End Sub

Private Sub silSOCKS5Read(ByVal wParam As Long)
  'V5
  
  Dim b() As Byte
  Dim bSize As Long
  Dim APIVal As Long
  Dim lSize As Long
  Dim reqRead As Long
  
  'Entered sub at wrong level
  If hS.Level < 2 Then Exit Sub
  
  Const SOCKS5_MethodPacketLen As Byte = 2
  Const SOCKS5_NoAuth As Byte = 0
  Const SOCKS5_UserPassVersion As Byte = 1
  Const SOCKS5_UserPass As Byte = 2
  Const SOCKS5_UserPassReply As Byte = 2
  Const SOCKS5_Version As Byte = 5
  Const SOCKS5_Reply As Byte = 10
  Const SOCKS5_NoMethods As Byte = 255
  
  'Get the amount to read
  ioctlsocket wParam, FIONREAD, lSize
  'If there's nothing to read, exit
  If lSize = 0 Then Exit Sub
  
  'Work out how much is valid to read
  Select Case hS.Level
    'Reply for supported methods
    Case 2
      With hData
        If lSize <> SOCKS5_MethodPacketLen Then
          'Haven't got expected 2 bytes ONLY
          If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_BadReplyForMethodV5, gErrStr(aeWinsockError_BadReplyForMethodV5)
          'Release socket
          ReleaseSocket False
          '
          Exit Sub
        End If
        'Set the required size
        reqRead = SOCKS5_MethodPacketLen
      End With
      
    Case 3
      With hData
        If lSize <> SOCKS5_UserPassReply Then
          'Haven't got 2 bytes
          If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_UPReplyTooBig, gErrStr(aeWinsockError_UPReplyTooBig)
          'close connection
          ReleaseSocket False
          Exit Sub
        End If
        'set the required size
        reqRead = SOCKS5_UserPassReply
      End With
      
    Case 4, 5
      With hData
        If lSize < SOCKS5_Reply Then
          'Raise error
          If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_SOCKS5ReplyToSmall, gErrStr(aeWinsockError_SOCKS5ReplyToSmall)
          'release
          ReleaseSocket False
          Exit Sub
        End If
      End With
      '
      reqRead = SOCKS5_Reply
     
  End Select
  
  'Resize the array to fit data
  ReDim b(0 To lSize - 1) As Byte
  'Get/save the size
  bSize = (UBound(b) - LBound(b) + 1)
  
  'Read the data from the socket
  APIVal = recv(wParam, b(0), bSize, MSG_NORMAL)
  '
  If APIVal = SOCKET_ERROR Then
    'There was an error
    If GetError(True) <> 0 Then
      ReleaseSocket False
    End If
    Exit Sub
  End If
  
  If APIVal < reqRead Then
    'Couldnt get data
    With hData
      If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_BadReadV5, gErrStr(aeWinsockError_BadReadV5)
    End With
    Exit Sub
  End If
  
  With hS
    Select Case .Level 'a
      '- SOCKS5 Method reply
      Case 2
        '-----
        ' 5 <Protocol Version>
        ' 0 <Method supported>
        '-----
        If b(0) <> SOCKS5_Version Then
          With hData
            'First byte wasn't 5
            If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_BadMethodReplyV5, gErrStr(aeWinsockError_BadMethodReplyV5)
            ReleaseSocket False
          End With
          Exit Sub
        End If
        
        'Get method
        Select Case b(1) 'b
        
          Case SOCKS5_NoAuth '0
            'No username/password needed
            silLevel 4
            'Send host info
            silSOCKS5SendHostInfo wParam
            Exit Sub
            
          Case SOCKS5_UserPass '2
            'Supports username/password for auth
            silSOCKS5UserPass wParam
            Exit Sub
            
          Case SOCKS5_NoMethods '255
            'No methods supported
            With hData
              If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_NoMethodsV5, gErrStr(aeWinsockError_NoMethodsV5)
            End With
            ReleaseSocket False
            Exit Sub
            
          Case Else
            'Unknown reply
            With hData
              If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_UnknownMethodV5, gErrStr(aeWinsockError_UnknownMethodV5)
            End With
            ReleaseSocket False
            Exit Sub
        End Select 'b
       '
       Exit Sub
       '
      
      Case 3
        '-----------
        ' 1 <version>
        ' 0 <valid> else failure
        '-----------
        If b(0) <> SOCKS5_UserPassVersion Then
          'version of u/p auth was incorrect
          With hData
            If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_BadUPVersionAuth, gErrStr(aeWinsockError_BadUPVersionAuth)
          End With
          ReleaseSocket False
          Exit Sub
        End If
        '
        Select Case b(1) 'c
          Case 0    'Success!
            '
            silLevel 4
            'Send host info
            silSOCKS5SendHostInfo wParam
            Exit Sub
          '
          Case Else 'Failed
            With hData
              If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_BadPasswordV5, gErrStr(aeWinsockError_BadPasswordV5)
            End With
            ReleaseSocket False
            Exit Sub
        End Select 'c
        '
        Exit Sub
        
    Case 4, 5
      '-------------------
      'SOCKS5 reply packet
      '-------------------
      ' 5 <version>      '0
      ' 0 <no error>     '1
      ' 0 <reserved>     '2
      ' 1 <address type> '3
      ' # <ip>           '4
      ' # <ip>           '5
      ' # <ip>           '6
      ' # <ip>           '7
      ' # <port>         '8
      ' # <port>         '9
      
      'Version byte in packet wasn't 5
      If b(0) <> SOCKS5_Version Then
        With hData
          If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_SOCKS5VersioninReplyInvalid, gErrStr(aeWinsockError_SOCKS5VersioninReplyInvalid)
        End With
        ReleaseSocket False
        Exit Sub
      End If
      '3rd byte in packet wasn't zero
      If b(2) <> 0 Then
        With hData
          If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_SOCKS5BadReplyNoRes, gErrStr(aeWinsockError_SOCKS5BadReplyNoRes)
        End With
        ReleaseSocket False
        Exit Sub
      End If
      '4th byte in packet, was not 1 (we don't support hostnames in reply packets)
      If b(3) <> 1 Then 'rare error
        With hData
          If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_SOCKSRepAddrTypeNoSupportByClient, gErrStr(aeWinsockError_SOCKSRepAddrTypeNoSupportByClient)
        End With
        ReleaseSocket False
        Exit Sub
      End If
      '
      Select Case b(1) 'd
        Case 0 'Success!
        Case 1 'general SOCKS server failure
          With hData
            If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_SOCKS5GeneralFail, gErrStr(aeWinsockError_SOCKS5GeneralFail)
          End With
          ReleaseSocket False
          Exit Sub
          
        Case 2 'connection not allowed by ruleset
          With hData
            If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_SOCKS5Rulesetfail, gErrStr(aeWinsockError_SOCKS5Rulesetfail)
          End With
          ReleaseSocket False
          Exit Sub
        
        Case 3 'Network unreachable
          With hData
            If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_SOCKS5NetUnreach, gErrStr(aeWinsockError_SOCKS5NetUnreach)
          End With
          ReleaseSocket False
          Exit Sub
        
        Case 4 'Host unreachable
          With hData
            If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_SOCKS5HostUnreach, gErrStr(aeWinsockError_SOCKS5HostUnreach)
          End With
          ReleaseSocket False
          Exit Sub
        
        Case 5 'Connection Refused
          With hData
            If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_SOCKS5ConnectionRefused, gErrStr(aeWinsockError_SOCKS5ConnectionRefused)
          End With
          ReleaseSocket False
          Exit Sub
        
        Case 6 'TTL Expired
          With hData
            If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_SOCKS5TTLExpired, gErrStr(aeWinsockError_SOCKS5TTLExpired)
          End With
          ReleaseSocket False
          Exit Sub
        
        Case 7 'Command not supported
          With hData
            If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_SOCKS5ComNoSupport, gErrStr(aeWinsockError_SOCKS5ComNoSupport)
          End With
          ReleaseSocket False
          Exit Sub
        
        Case 8 'Address type not supported
          With hData
            If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_SOCKS5AddrTypeNotSupported, gErrStr(aeWinsockError_SOCKS5AddrTypeNotSupported)
          End With
          ReleaseSocket False
          Exit Sub
        
        Case 9 To 255 'unknown error
          With hData
            If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_SOCKS5UnAssignedError, gErrStr(aeWinsockError_SOCKS5UnAssignedError)
          End With
          ReleaseSocket False
          Exit Sub
        
      End Select 'd
      '
      Dim hIP As String
      Dim lPort As Long
      '
      ' Get port information
      '
      'Copy last byte (9th) into first byte in a long
      CopyMemory lPort, b(9), 1
      'Copy 8th byte into the 2nd byte in a long
      CopyMemory ByVal VarPtr(lPort) + 1, b(8), 1
      '
      ' Get IP4 address  (4-7)
      '
      hIP = CStr(b(4) & "." & b(5) & "." & b(6) & "." & b(7))
      '
      Select Case .Level ' #u
        '//--
        Case 4
        '//--
        With hData 't
          If hS.Meth = SOCKSConnect Then
          '------C O N N E C T E D------
            With hS
              'Save server bound address
              .SeenIP = hIP
              'Save serveer bound port
              .SeenPort = lPort
            End With
            'Add old interface relation
            SetInterface .hSocket, Me, "SCK"
            'Delete SOCKS relation
            DeleteInterface .hSocket, "SOCKS"
            'Reenable messages
            WSAAsyncSelect .hSocket, ws_hwnd, SCK_REPLY, FD_ALL
            Exit Sub 'Exit
          '-----------------------------
          Else
          '---------B I N D E D---------
            With hS
              'Save server ip
              .BindHost = hIP
              'Save server bound port
              .lBindPort = lPort
            End With
            'Change level
            silLevel 5
            'Server socket is listening
            SetState SOCKS_Listening
            'Save client IP
            .RemoteIP = hS.ThatIP
            'Save client port
            .RemotePort = hS.ThatPort
            'Raise event
            If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Binded lPort, True
            Exit Sub ' Exit
          '-----------------------------
          End If
        End With 't
        '
        '//--
        Case 5
        '//--
          'Save indirect host IP
          .IndirectHost = hIP
          'Save indirect port
          .IndirectPort = lPort
          '
          With hData
            'Reset old relation
            SetInterface .hSocket, Me, "SCK"
            'Delete SOCKS relation
            DeleteInterface .hSocket, "SOCKS"
            'Renable messages for SCK_REPLY
            WSAAsyncSelect .hSocket, ws_hwnd, SCK_REPLY, FD_ALL
            '
            Exit Sub '
          End With
        End Select '#u
        '
    End Select 'a
  End With
   
End Sub

Private Sub silSOCKS5UserPass(ByVal wParam As Long)
  'V5
  Dim UP As String
  Dim b() As Byte
  Dim bSize As Long
  Dim APIVal As Long
  '
  With hS
    'Wrong level
    If .Level <> 2 Then Exit Sub
    'Use username/password authentication V1
    UP = Chr(1)
    'The length of the username
    UP = UP & Chr(Len(.UserName))
    'The username
    UP = UP & .UserName
    'The length of the password
    UP = UP & Chr(Len(.PassWord))
    'the password
    UP = UP & .PassWord
  End With
  'Change the level
  silLevel 3
  'resize the array
  ReDim b(0 To Len(UP) - 1) As Byte
  'Convert the UNICODE string to a byte array
  b = StrConv(UP, vbFromUnicode)
  'Get the size of b
  bSize = (UBound(b) - LBound(b) + 1)
  'Send to SOCKS
  APIVal = send(wParam, b(0), bSize, MSG_NORMAL)
  If APIVal = SOCKET_ERROR Then
    If GetError(True) <> 0 Then
      ReleaseSocket False
    End If
    Exit Sub 'Exit
  End If
  'Couldn't send all of it
  If APIVal <> bSize Then
    With hData
      If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_BadSendUPV5, gErrStr(aeWinsockError_BadSendUPV5)
      ReleaseSocket False
    End With
    Exit Sub 'Exit
  End If
End Sub

Private Sub silSOCKS5SendHostInfo(ByVal wParam As Long)
  'V5
  Dim APIVal As Long
  Dim b() As Byte
  Dim bSize As Long
  Dim INet As Inetaddress
  Dim bPort As Byte
  Dim bPortEx As Byte
  Dim S5 As String
  
  With hS
    'Not the right level
    If .Level <> 4 Then Exit Sub
    '--------------
    'SOCKS5 request
    '--------------
    ' 5   <protocol version>
    ' 1   <command, connect, bind>
    ' 0   <reserved>
    ' 1   <address type, IP4, Hostname, IP6>
    ' 127 <IP> Start of dst addr
    ' 0   <IP>
    ' 0   <IP
    ' 1   <IP> End of dst addr
    ' 26  <Port> (this is 2 bytes of a DWORD)
    ' 11  <Port>
    'Using SOCKS protocol 5
    S5 = Chr(5)
    'set CONNECT/BIND byte
    If .Meth = SOCKSConnect Then
      S5 = S5 & Chr(1) 'Connect through
    Else
      S5 = S5 & Chr(2) 'Bind on
    End If
    'Reserved byte
    S5 = S5 & Chr(0)
    'Using IP4 address type (01)
    S5 = S5 & Chr(1)
    '
    'Copy 'thatIP' long into a struct with 4 byte members
    '
    CopyMemory INet, inet_addr(.ThatIP), Len(INet)
    'Copy each byte into the string
    With INet 'Inet
      S5 = S5 & Chr(.Byte4) & Chr(.Byte3) & Chr(.Byte2) & Chr(.Byte1)
    End With 'Inet
    '
    'Copy port information
    '
    CopyMemory bPort, ByVal VarPtr(.ThatPort) + 1, 1
    'Read 1st byte
    CopyMemory bPortEx, .ThatPort, 1
    'Save in packet
    S5 = S5 & Chr(bPort) & Chr(bPortEx)
    'Resize b()
    ReDim b(0 To Len(S5) - 1) As Byte
    'Turn UNICODE string into bytearray
    b = StrConv(S5, vbFromUnicode)
    'Save size of b()
    bSize = (UBound(b) - LBound(b) + 1)
    'send to server
    APIVal = send(wParam, b(0), bSize, MSG_NORMAL)
    If APIVal = SOCKET_ERROR Then
      If GetError(True) <> 0 Then
        ReleaseSocket False
      End If
      Exit Sub
    End If
    'Didn't send entire packet
    If APIVal <> bSize Then
      With hData
        If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_BadSendOfHostInfo, gErrStr(aeWinsockError_BadSendOfHostInfo)
      End With
      ReleaseSocket False
      Exit Sub
    End If
    '
  End With
End Sub

Private Sub silSOCKS5Close(ByVal wParam As Long)
  '1 + asking for method
  '2 + got method reply
  '3 + username/password stage.. no u/p goto level 4
  '4 + Send connect/bind request
  '4 + got a connect or bind (after send)
  '5 + connected from listen
  With hData
    Select Case hS.Level
      Case 1
        If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_SOCKSCloseLevel1, gErrStr(aeWinsockError_SOCKSCloseLevel1)
      Case 2
        If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_SOCKSCloseLevel2, gErrStr(aeWinsockError_SOCKSCloseLevel2)
      Case 3
        If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_SOCKSCloseLevel3, gErrStr(aeWinsockError_SOCKSCloseLevel3)
      Case 4
        If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_SOCKSCloseLevel4, gErrStr(aeWinsockError_SOCKSCloseLevel4)
      Case 5
        If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_SOCKSCloseLevel5, gErrStr(aeWinsockError_SOCKSCloseLevel5)
    End Select
  End With
  'This sub was invoked by an FD_CLOSE event therefore
  'it was signalled from the remote side
  ReleaseSocket True
End Sub

Private Sub silSOCKS4Connect(ByVal wParam As Long)

  'V4
  Dim b() As Byte
  Dim bSize As Long
  Dim S4 As String
  Dim bPort As Byte
  Dim bPortEx As Byte
  Dim INet As Inetaddress
  Dim APIVal As Long
 
        With hS
        'Stop neg timeouut
        SetProp ws_hwnd, CStr(hData.hClassPtr & "#ntimeout"), 1
        'Don't resend
        If .Level > 2 Then Exit Sub
          'Build a SOCKS4 Connect/Bind packet
          S4 = Chr(4)
            If .Meth = SOCKSConnect Then
              'Want connection
              S4 = S4 & Chr(1)
            Else
              'Want Bind
              S4 = S4 & Chr(2)
            End If
          '
          'Work out port information
          
          'Read 2nd byte (goes first)
          CopyMemory bPort, ByVal VarPtr(.ThatPort) + 1, 1
          'Read 1 byte
          CopyMemory bPortEx, .ThatPort, 1
          'Save in packet
          S4 = S4 & Chr(bPort) & Chr(bPortEx)
          'Copy IP4 Address into 4 bytes
          CopyMemory INet, inet_addr(.ThatIP), Len(INet)
          'Save in packet
          With INet
            S4 = S4 & Chr(.Byte4) & Chr(.Byte3) & Chr(.Byte2) & Chr(.Byte1)
          End With
          'If a username is given, include it.
          If Len(.UserName) <> 0 Then S4 = S4 & .UserName
          'Null term the packet
          S4 = S4 & Chr(0)
          'Resize b() to fit 'S4'
          ReDim b(0 To Len(S4) - 1) As Byte
          'Convert the UNICODE string into a byte array
          b = StrConv(S4, vbFromUnicode)
          'Save the size of b
          bSize = (UBound(b) - LBound(b) + 1)
          'Send b() to server
          'Change level
          silLevel 2
          '
          With hData
            APIVal = send(.hSocket, b(0), bSize, MSG_NORMAL)
            If APIVal = SOCKET_ERROR Then
              'Error
              If GetError(True) <> 0 Then
                'Releasesocket if error code not 'would block'
                ReleaseSocket False
              End If
              'exit
              Exit Sub
            End If
            'If sent length doesn't match packet size error out
            If APIVal <> bSize Then
              If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_CantSendPacketV4, gErrStr(aeWinsockError_CantSendPacketV4)
              ReleaseSocket False
            End If
          End With
          
      End With
End Sub

Private Sub silSOCKS4Read(ByVal wParam As Long)
 
  ' When a SOCKS4 server replies, it sends back
  ' an 8 byte packet
  
  Const SOCKS4ReplyLen As Byte = 8
  
  Dim b() As Byte
  Dim bSize As Long
  Dim lSize As Long
  Dim APIVal As Long
  Dim lPort As Long
  Dim hIP As String
  
  'Get the amount of data to read
  ioctlsocket wParam, FIONREAD, lSize
  'No data to read!
  If lSize = 0 Then Exit Sub
  
  If hS.Level = 1 Then
    'If we read before we've sent, the server isn't a SOCKS4
    With hData
      If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_NotSOCKS4, gErrStr(aeWinsockError_NotSOCKS4)
      ReleaseSocket False
    End With
    Exit Sub
  End If
  
  With hS 'hs
    'Stop timeout
    SetProp ws_hwnd, CStr(hData.hClassPtr & "#ntimeout"), 1
    'Turn off read
    WSAAsyncSelect wParam, ws_hwnd, SOCKS_REPLY, FD_ALL And Not FD_READ
    'Got a reply, still may not be a SOCKS4 server
    If lSize <> SOCKS4ReplyLen Then
      'Ain't SOCKS4
      With hData
        If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_NotSOCKS4, gErrStr(aeWinsockError_NotSOCKS4)
        ReleaseSocket False
      End With
      Exit Sub
    End If
    
    'Get the 8 bytes
    ReDim b(0 To 7) As Byte
    APIVal = recv(wParam, b(0), 8, 0)
    If APIVal = SOCKET_ERROR Then
      'Didn't get the data
      If GetError(True) <> 0 Then
        ReleaseSocket False
        Exit Sub
      End If
    End If
    If APIVal <> SOCKS4ReplyLen Then
      'Didn't read all 8 bytes
      With hData
        If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_CantReadPacketV4, gErrStr(aeWinsockError_CantReadPacketV4)
      End With
        ReleaseSocket False
      Exit Sub
    End If
      '----------------------
      'Valid reply
      '----------------------
      '0   - Always zero
      '90  - Result byte
      '4   - Port
      '43  - Port
      '127 - IP
      '0   - IP
      '0   - IP
      '1   - IP
      '----------------------
      
      'First byte wasn't zero as expected with SOCKS4
      If b(0) <> 0 Then
        With hData
          If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_NotSOCKS4Reply, gErrStr(aeWinsockError_NotSOCKS4Reply)
          ReleaseSocket False
        End With
        Exit Sub
      End If
      'Second byte has error information or success!
      Select Case b(1)
        Case 90
          'Success
         
        Case 91
          'Rejected or failed
          With hData
            If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_RejectedOrFailV4, gErrStr(aeWinsockError_RejectedOrFailV4)
          End With
          ReleaseSocket False
          Exit Sub
          
        Case 92
          'Can't connect to IDENTD
          With hData
            If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_CantConnectToIdentd, gErrStr(aeWinsockError_CantConnectToIdentd)
          End With
          ReleaseSocket False
          Exit Sub
          
        Case 93
          'IDENTD username and passed username don't match
          With hData
            If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_UserNameIdentNonMatch, gErrStr(aeWinsockError_UserNameIdentNonMatch)
          End With
          ReleaseSocket False
          Exit Sub
          
        Case Else
          'Unknown reply
          With hData
            If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_ErrorUnknownV4, gErrStr(aeWinsockError_ErrorUnknownV4)
          End With
          ReleaseSocket False
          Exit Sub
          
        End Select
        '
        'Get port used by SOCKS4 to make the connection
        '
        'Copy 3rd byte in b() to first byte in long
        CopyMemory lPort, b(3), 1
        'Copy 2nd byte in b() to second byte in long
        CopyMemory ByVal VarPtr(lPort) + 1, b(2), 1
        
        'Make sure port isn't invalid
        If lPort = 0 Then
          With hData
            If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_InvalidPortV4, gErrStr(aeWinsockError_InvalidPortV4)
            ReleaseSocket False
          End With
          Exit Sub
        End If
        
        ' Get the IP address
        ' This IP maybe different to the proxy IP
        ' As it may connect to the passed host on another network connection
        
        hIP = b(4) & "." & b(5) & "." & b(6) & "." & b(7)
        '
  Select Case .Level
    Case 2
        With hData 'a
        If hS.Meth = SOCKSConnect Then
          With hS
            .SeenIP = hIP
            .SeenPort = lPort
          End With
          'Add the normal interface again
          SetInterface .hSocket, Me, "SCK"
          'Delete the SOCKS interface
          DeleteInterface .hSocket, "SOCKS"
          'Reenable messages for standard socket code
          WSAAsyncSelect .hSocket, ws_hwnd, SCK_REPLY, FD_ALL
          '
          Exit Sub
        Else
          'Reenable winsock messages
          WSAAsyncSelect .hSocket, ws_hwnd, SOCKS_REPLY, FD_ALL
          With hS
            'Save the IP
            .BindHost = hIP
            'Save the random binded port
            .lBindPort = lPort
          End With
          'Change level
          silLevel 4
          'invoke SOCKS listen.. state
          SetState SOCKS_Listening
          'Save client IP in remote host
          'Even though we are connected to the proxy
          .RemoteIP = hS.ThatIP
          'Save port information of client
          .RemotePort = hS.ThatPort
          'Raise bind
          If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Binded lPort, True
        End If
        End With 'a
        
        Case 3
        'Do nothing
        Exit Sub
        
        Case 4
          'They connected
          With hS
            'Save the host of the indirect party
            .IndirectHost = hIP
            'And the port
            .IndirectPort = lPort
          End With
          With hData
            'Add the normal interface again
            SetInterface .hSocket, Me, "SCK"
            'Delete the SOCKS interface
            DeleteInterface .hSocket, "SOCKS"
            'Reenable messages for standard socket code
            WSAAsyncSelect .hSocket, ws_hwnd, SCK_REPLY, FD_ALL
          End With
          Exit Sub
    End Select
  End With 'hs

End Sub

Private Sub silSOCKS4Close(ByVal wParam As Long)

  Select Case hS.Level
    Case 1

    Case 2 'Closed no reply
      'The connection was closed AFTER we sent a connection request
      'probably because the port was refused
      With hData
        If .hEventPtr <> 0 Then EvtFace(.hEventPtr).Error aeWinsockError_NoErrorReplyV4, gErrStr(aeWinsockError_NoErrorReplyV4)
        ReleaseSocket True
      End With
      Exit Sub
      
    Case Else
      'Had successfully got a BIND
      'and was waiting for incoming, but the connection closed
      'binded remote port, lost
      With hData
        If .hEventPtr Then EvtFace(.hEventPtr).Error aeWinsockError_CloseAfterBindV4, gErrStr(aeWinsockError_CloseAfterBindV4)
      End With
      ReleaseSocket True
      Exit Sub
  End Select
End Sub

Public Sub FlushBuffer()
  With hBuf
    If .DontWrite = False Then
      'Stop writing
      .DontWrite = True
      'Free object if it exists
      If .hMem <> 0 Then
        'Free memory object
        GlobalFree .hMem
        'Reset handle
        .hMem = 0
        .Size = 0
        .ActualTotalSize = 0
      End If
      'Allow writing
      .DontWrite = False
    End If
  End With
End Sub

'Returns:
' 0 := Memory buffer doesn't exist
' / := Size of buffer therefore it exists
Public Function BufferExists() As Long
  BufferExists = hBuf.Size
End Function

Public Property Get UseSOCKS() As Boolean
  With hS
    UseSOCKS = .bUseSOCKS
  End With
End Property
'
'Note:
'     If this is set to true, and SOCKSHost/SOCKSPort
'     aren't filled it will be turned off
'
Public Property Let UseSOCKS(ByVal vData As Boolean)
  With hS
    If hData.hSocket <> SOCKET_ERROR Then Exit Property
    'Don't allow if already connected
    If .Level <> 0 Then Exit Property
    If .bUseSOCKS <> vData Then
      .bUseSOCKS = vData
    End If
  End With
End Property

Public Property Get SOCKSHost() As String
  SOCKSHost = hS.ProxyName
End Property
Public Property Let SOCKSHost( _
  ByVal vData As String)
  With hS
    'Exit if connected
    If .Level <> 0 Then Exit Property
    If .ProxyName <> vData Then
      'Only set if it's not the same
      .ProxyName = vData
      'if 'proxyname' is an I.P. address save it
      If IsHost(vData) = False Then
        .ProxyIP = vData
        Else
        .ProxyIP = ""
      End If
    End If
  End With
End Property

Public Property Get SOCKSHostIP() As String
  SOCKSHostIP = hS.ProxyIP
End Property

Public Property Get SOCKSPort() As Long
  SOCKSPort = hS.ProxyPort
End Property
Public Property Let SOCKSPort(ByVal vData As Long)
  With hS
    If .Level <> 0 Then Exit Property
    If vData < 1 Then vData = 1
    If .ProxyPort <> vData Then
      .ProxyPort = vData
    End If
  End With
End Property

Public Property Get SOCKSProtocol() As aeSOCKSProtocol
  SOCKSProtocol = hS.TOC
End Property
Public Property Let SOCKSProtocol(ByVal vData As aeSOCKSProtocol)
  With hS
    If .Level <> 0 Then Exit Property
    If .TOC <> vData Then .TOC = vData
  End With
End Property

Public Property Get SOCKSMethod() As aeSOCKSMethod
  SOCKSMethod = hS.Meth
End Property
Public Property Let SOCKSMethod(ByVal vData As aeSOCKSMethod)
  With hS
    If .Level <> 0 Then Exit Property
    If .Meth <> vData Then .Meth = vData
  End With
End Property

Public Property Get SOCKSTimeout() As Long
  SOCKSTimeout = hS.NegTimeOut
End Property
Public Property Let SOCKSTimeout(ByVal vData As Long)
  With hS
    'Allow to be set at any level
    If vData < 0 Then vData = 0
    If .NegTimeOut <> vData Then .NegTimeOut = vData
    With hData
      SetProp ws_hwnd, CStr(.hClassPtr & "#negcount"), 0
    End With
  End With
End Property

Public Property Get SOCKSUserName() As String
  SOCKSUserName = hS.UserName
End Property
Public Property Let SOCKSUserName( _
ByVal vData As String)
  With hS
    If .Level <> 0 Then Exit Property
    If .UserName <> vData Then .UserName = vData
    .bUP = Len(.UserName) <> 0 And Len(.PassWord) <> 0
  End With
End Property

Public Property Get SOCKSPassWord() As String
  SOCKSPassWord = hS.PassWord
End Property
Public Property Let SOCKSPassWord(ByVal vData As String)
  With hS
    If .Level <> 0 Then Exit Property
    If .PassWord <> vData Then .PassWord = vData
    .bUP = Len(.UserName) <> 0 And Len(.PassWord) <> 0
  End With
End Property

Public Property Get SOCKSBindedIP() As String
  SOCKSBindedIP = hS.BindHost
End Property
Public Property Get SOCKSBindedPort() As Long
  SOCKSBindedPort = hS.lBindPort
End Property

Public Property Get SOCKSSeenIP() As String
  SOCKSSeenIP = hS.SeenIP
End Property
Public Property Get SOCKSSeenPort() As Long
  SOCKSSeenPort = hS.SeenPort
End Property

Public Property Get SOCKSIndirectHost() As String
  SOCKSIndirectHost = hS.IndirectHost
End Property
Public Property Get SOCKSIndirectPort() As Long
  SOCKSIndirectPort = hS.IndirectPort
End Property

'Returns:
'
' * True  := The contents of the memory buffer has been written to disk
' * False := failed
'
'Inputs:
' * PathAndFileName := The path and file name to create the file e.g. c:\buffer.txt
' * [bOverWrite] := If true, if the file exists it's overwritten
'
'Note:
'This function doesn't REMOVE the memory buffer
'
Public Function SaveBufferToFile( _
  ByVal PathAndFileName As String, _
  Optional bOverWrite As Boolean = False) As Boolean
  
  'Path/Filename was nothing
  If Len(PathAndFileName) = 0 Then Exit Function
  'Memory object doesn't exist
  If hBuf.hMem = 0 Then Exit Function
  
  Dim lPtr As Long
  Dim hFile As Long
  Dim bFlag As Long
  Dim bWriteLen As Long
  Dim APIVal As Long
  
  If bOverWrite = True Then
    'Over write the file
    bFlag = CREATE_ALWAYS
  Else
    'Open if it doesn't exist
    bFlag = CREATE_NEW
  End If
  
  'Get a handle to the file
  hFile = CreateFile(PathAndFileName, GENERIC_WRITE, 0, ByVal 0&, bFlag, FILE_ATTRIBUTE_NORMAL Or FILE_ATTRIBUTE_ARCHIVE, ByVal 0&)
  'Couldn't get handle
  If hFile = SOCKET_ERROR Then Exit Function
  
  With hBuf
    If .DontWrite = False Then
      'Stop writing
      .DontWrite = True
      'Get the lock pointer
      lPtr = GlobalLock(.hMem)
      If lPtr Then
        'Write to disk
        APIVal = WriteFile(hFile, ByVal lPtr, .Size, bWriteLen, ByVal 0&)
        'Wrote memory to disk
        SaveBufferToFile = APIVal
        'Unlock the memory object
        GlobalUnlock .hMem
        'Null pointer
        lPtr = 0
      End If
      'Allow writing
      .DontWrite = False
    End If
  End With
  
  'Close the file
  CloseHandle hFile
End Function
